# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
import os

import pytest

from numpy.testing import (assert_equal, assert_almost_equal)

import MDAnalysis as mda
from MDAnalysis.transformations import translate
from MDAnalysisTests.datafiles import (PDB, PSF, CRD, DCD,
                                       GRO, XTC, TRR, PDB_small, PDB_closed)
from MDAnalysisTests.util import no_warning


class TestChainReader(object):
    prec = 3

    @pytest.fixture()
    def universe(self):
        return mda.Universe(PSF,
                            [DCD, CRD, DCD, CRD, DCD, CRD, CRD])

    @pytest.fixture()
    def transformed(ref):
        return mda.Universe(PSF,
                            [DCD, CRD, DCD, CRD, DCD, CRD, CRD],
                            transformations=[translate([10,10,10])])
                            
    def test_regular_repr(self):
        u = mda.Universe(PSF, [DCD, CRD, DCD])
        assert_equal("<ChainReader containing adk_dims.dcd, adk_open.crd, adk_dims.dcd with 197 frames of 3341 atoms>", u.trajectory.__repr__())
        
                                
    def test_truncated_repr(self, universe):
        assert_equal("<ChainReader containing adk_dims.dcd and 6 more with 298 frames of 3341 atoms>", universe.trajectory.__repr__())

    def test_next_trajectory(self, universe):
        universe.trajectory.rewind()
        universe.trajectory.next()
        assert_equal(universe.trajectory.ts.frame, 1, "loading frame 2")

    def test_n_atoms(self, universe):
        assert_equal(universe.trajectory.n_atoms, 3341,
                     "wrong number of atoms")

    def test_n_frames(self, universe):
        assert_equal(universe.trajectory.n_frames, 3 * 98 + 4,
                     "wrong number of frames in chained dcd")

    def test_iteration(self, universe):
        for ts in universe.trajectory:
            pass  # just forward to last frame
        assert_equal(universe.trajectory.n_frames - 1, ts.frame,
                     "iteration yielded wrong number of frames ({0:d}), "
                     "should be {1:d}".format(ts.frame,
                                              universe.trajectory.n_frames))

    def test_jump_lastframe_trajectory(self, universe):
        universe.trajectory[-1]
        assert_equal(universe.trajectory.ts.frame + 1,
                     universe.trajectory.n_frames,
                     "indexing last frame with trajectory[-1]")

    def test_slice_trajectory(self, universe):
        frames = [ts.frame for ts in universe.trajectory[5:17:3]]
        assert_equal(frames, [5, 8, 11, 14], "slicing dcd [5:17:3]")

    def test_full_slice(self, universe):
        trj_iter = universe.trajectory[:]
        frames = [ts.frame for ts in trj_iter]
        assert_equal(frames, np.arange(universe.trajectory.n_frames))

    def test_frame_numbering(self, universe):
        universe.trajectory[98]  # index is 0-based and frames are 0-based
        assert_equal(universe.trajectory.frame, 98, "wrong frame number")

    def test_next_after_frame_numbering(self, universe):
        universe.trajectory[98]  # index is 0-based and frames are 0-based
        universe.trajectory.next()
        assert_equal(universe.trajectory.frame, 99, "wrong frame number")

    def test_frame(self, universe):
        universe.trajectory[0]
        coord0 = universe.atoms.positions.copy()
        # forward to frame where we repeat original dcd again:
        # dcd:0..97 crd:98 dcd:99..196
        universe.trajectory[99]
        assert_equal(universe.atoms.positions, coord0,
                     "coordinates at frame 1 and 100 should be the same!")

    def test_time(self, universe):
        universe.trajectory[30]  # index and frames 0-based
        assert_almost_equal(
            universe.trajectory.time, 30.0, 5, err_msg="Wrong time of frame")

    def test_write_dcd(self, universe, tmpdir):
        """test that ChainReader written dcd (containing crds) is correct
        (Issue 81)"""
        outfile = str(tmpdir) + "chain-reader.dcd"
        with mda.Writer(outfile, universe.atoms.n_atoms) as W:
            for ts in universe.trajectory:
                W.write(universe)
        universe.trajectory.rewind()
        u = mda.Universe(PSF, outfile)
        for (ts_orig, ts_new) in zip(universe.trajectory, u.trajectory):
            assert_almost_equal(
                ts_orig._pos,
                ts_new._pos,
                self.prec,
                err_msg="Coordinates disagree at frame {0:d}".format(
                    ts_orig.frame))       
    
    def test_transform_iteration(self, universe, transformed):
        vector = np.float32([10,10,10])
        # # Are the transformations applied and
        # are the coordinates "overtransformed"?
        # iterate once:
        for ts in transformed.trajectory:
            frame = ts.frame
            ref = universe.trajectory[frame].positions + vector
            assert_almost_equal(ts.positions, ref, decimal = 6)
        # iterate again:
        for ts in transformed.trajectory:
            frame = ts.frame
            ref = universe.trajectory[frame].positions + vector
            assert_almost_equal(ts.positions, ref, decimal = 6)
    
    def test_transform_slice(self, universe, transformed):
        vector = np.float32([10,10,10])
        # what happens when we slice the trajectory?
        for ts in transformed.trajectory[5:17:3]:
            frame = ts.frame
            ref = universe.trajectory[frame].positions + vector
            assert_almost_equal(ts.positions, ref, decimal = 6)
    
    def test_transform_switch(self, universe, transformed):
        vector = np.float32([10,10,10])
        # grab a frame:
        ref = universe.trajectory[2].positions + vector
        assert_almost_equal(transformed.trajectory[2].positions, ref, decimal = 6)
        # now switch to another frame
        newref = universe.trajectory[10].positions + vector
        assert_almost_equal(transformed.trajectory[10].positions, newref, decimal = 6)
        # what happens when we comeback to the previous frame?
        assert_almost_equal(transformed.trajectory[2].positions, ref, decimal = 6)
    
    def test_transfrom_rewind(self, universe, transformed):
        vector = np.float32([10,10,10])
        ref = universe.trajectory[0].positions + vector
        transformed.trajectory.rewind()
        assert_almost_equal(transformed.trajectory.ts.positions, ref, decimal = 6)

class TestChainReaderCommonDt(object):
    common_dt = 100.0
    prec = 3

    @pytest.fixture()
    def trajectory(self):
        universe = mda.Universe(
            PSF, [DCD, CRD, DCD, CRD, DCD, CRD, CRD], dt=self.common_dt)
        return universe.trajectory

    def test_time(self, trajectory):
        # We test this for the beginning, middle and end of the trajectory.
        for frame_n in (0, trajectory.n_frames // 2, -1):
            trajectory[frame_n]
            assert_almost_equal(
                trajectory.time,
                trajectory.frame * self.common_dt,
                5,
                err_msg="Wrong time for frame {0:d}".format(frame_n))


class TestChainReaderFormats(object):
    """Test of ChainReader with explicit formats (Issue 76)."""

    def test_set_all_format_tuples(self):
        universe = mda.Universe(GRO, [(PDB, 'pdb'), (XTC, 'xtc'), (TRR,
                                                                   'trr')])
        assert universe.trajectory.n_frames == 21
        assert_equal(universe.trajectory.filenames, [PDB, XTC, TRR])

    def test_set_format_tuples_and_format(self):
        universe = mda.Universe(GRO, [(PDB, 'pdb'), GRO, GRO, (XTC, 'xtc'), 
                                      (TRR, 'trr')], format='gro')
        assert universe.trajectory.n_frames == 23
        assert_equal(universe.trajectory.filenames, [PDB, GRO, GRO, XTC, TRR])
        
        with pytest.raises(TypeError) as errinfo:
            mda.Universe(GRO, [(PDB, 'pdb'), GRO, GRO, (XTC, 'xtc'), 
                                      (TRR, 'trr')], format='pdb')
        assert 'Unable to read' in str(errinfo.value)


    def test_set_one_format_tuple(self):
        universe = mda.Universe(PSF, [(PDB_small, 'pdb'), DCD])
        assert universe.trajectory.n_frames == 99

    def test_set_all_formats(self):
        with pytest.raises(TypeError) as errinfo:
            mda.Universe(PDB, [PDB, GRO], format='gro')
        assert 'Unable to read' in str(errinfo.value)

        universe = mda.Universe(GRO, [PDB, PDB, PDB], format='pdb')
        assert_equal(universe.trajectory.filenames, [PDB, PDB, PDB])


def build_trajectories(folder, sequences, fmt='xtc'):
    """
    A scenario is given as a series of time sequences. The result is
    returned as a list of time and origin of the frame. Each element of that
    result list correspond to a frame of the conatenated trajectory, The
    first element of each tuple is the time it corresponds to, the second
    element is the index of the input the frame comes from.

    See original gist and PR 1728 for comparison with gromacs trjcat.

    https://gist.github.com/jbarnoud/cacd0957d3df01d1577f640b20e86039
    """
    template = 'trjcat_test_{{}}.{}'.format(fmt)
    template = os.path.join(folder, template)

    # Use an empty universe to have a topology
    utop = mda.Universe.empty(1, trajectory=True)

    # Create synthetic trajectories. The times come from the user input,
    # the coordinates indicate the index of the input.
    fnames = []
    for index, subseq in enumerate(sequences):
        coords = np.zeros((len(subseq), 1, 3), dtype=np.float32) + index
        u = mda.Universe(utop._topology, coords)
        out_traj = mda.Writer(template.format(index), n_atoms=len(u.atoms))
        fnames.append(out_traj.filename)
        with out_traj:
            for ts, time in zip(u.trajectory, subseq):
                # The time step needs a box to avoid a warning
                ts.dimensions = [10, 10, 10, 90, 90, 90]
                # The time is set from the user input
                ts.time = time
                out_traj.write(u)
    return utop, fnames


class TestChainReaderContinuous(object):
    class SequenceInfo(object):
        def __init__(self, seq, n_frames, order):
            self.seq = seq
            self.n_frames = n_frames
            self.order = order

    @pytest.mark.parametrize('fmt', ('xtc', 'trr'))
    @pytest.mark.parametrize('seq_info',[
        SequenceInfo(seq=([0, 1, 2, 3], [2, 3, 4, 5], [4, 5, 6, 7]),
                     n_frames=8,
                     order=[0, 0, 1, 1, 2, 2, 2, 2]),
        SequenceInfo(seq=([0, 1, 2, 4], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
                     n_frames=10,
                     order=np.ones(10)),
        SequenceInfo(seq=([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [0, 1, 2, 3]),
                     n_frames=4,
                     order=np.ones(4)),
        SequenceInfo(seq=([5, 6, 7, 8, 9], [2, 3, 4, 5, 6], [0, 1, 2, 3]),
                     n_frames=10,
                     order=[2, 2, 1, 1, 1, 0, 0, 0, 0, 0]),
        SequenceInfo(seq=([0, 1, 2],) * 3,
                     n_frames=3,
                     order=[2, 2, 2]),
        SequenceInfo(seq=([0, 1, 2, 3,], [3, 4], [4, 5, 6, 7]),
                     n_frames=8,
                     order=[0, 0, 0, 1, 2, 2, 2, 2]),
        SequenceInfo(seq=([5, 6, 7, 8, 9], [2, 3, 4, 5, 6], [0, 1, 2, 3]),
                     n_frames=10,
                     order=[2, 2, 1, 1, 1, 0, 0, 0, 0, 0]),
        SequenceInfo(seq=[list(range(0, 6)), list(range(2, 5)),
                          list(range(2, 5)), list(range(2, 5)),
                          list(range(3, 8))],
                     n_frames=8,
                     order=[0, 0, 3, 4, 4, 4, 4, 4]),
    ])
    def test_order(self, seq_info, tmpdir, fmt):
        folder = str(tmpdir)
        utop, fnames = build_trajectories(folder, sequences=seq_info.seq, fmt=fmt)
        u = mda.Universe(utop._topology, fnames, continuous=True)
        assert u.trajectory.n_frames == seq_info.n_frames
        for i, ts in enumerate(u.trajectory):
            assert_almost_equal(i, ts.time, decimal=4)
            # check we have used the right trajectory
            assert seq_info.order[i] == int(ts.positions[0, 0])

    def test_start_frames(self, tmpdir):
        folder = str(tmpdir)
        sequences = ([0, 1, 2, 3], [2, 3, 4, 5], [4, 5, 6, 7])
        utop, fnames = build_trajectories(folder, sequences=sequences,)
        u = mda.Universe(utop._topology, fnames, continuous=True)
        assert_equal(u.trajectory._start_frames, [0, 2, 4])

    def test_missing(self, tmpdir):
        folder = str(tmpdir)
        sequences = ([0, 1, 2, 3], [5, 6, 7, 8, 9])
        utop, fnames = build_trajectories(folder, sequences=sequences,)
        u = mda.Universe(utop._topology, fnames, continuous=True)
        assert u.trajectory.n_frames == 9

    def test_warning(self, tmpdir):
        folder = str(tmpdir)
        # this sequence *should* trigger a warning
        sequences = ([0, 1, 2, 3], [5, 6, 7])
        utop, fnames = build_trajectories(folder, sequences=sequences,)
        with pytest.warns(UserWarning):
            mda.Universe(utop._topology, fnames, continuous=True)

    def test_interleaving_error(self, tmpdir):
        folder = str(tmpdir)
        # interleaving is not supported by chainreader
        sequences = ([0, 2, 4, 6], [1, 3, 5, 7])
        utop, fnames = build_trajectories(folder, sequences=sequences,)
        with pytest.raises(RuntimeError):
            mda.Universe(utop._topology, fnames, continuous=True)

    def test_easy_trigger_warning(self, tmpdir):
        folder = str(tmpdir)
        # this sequence shouldn't trigger a warning
        sequences = ([0, 1, 2, 3], [2, 3, 4, 5], [4, 5, 6, 7])
        utop, fnames = build_trajectories(folder, sequences=sequences,)
        with no_warning(UserWarning):
            mda.Universe(utop._topology, fnames, continuous=True)

    def test_single_frames(self, tmpdir):
        folder = str(tmpdir)
        sequences = ([0, 1, 2, 3], [5, ])
        utop, fnames = build_trajectories(folder, sequences=sequences,)
        with pytest.raises(RuntimeError):
            mda.Universe(utop._topology, fnames, continuous=True)

    def test_mixed_filetypes(self):
        with pytest.raises(ValueError):
            mda.Universe(PDB, [XTC, TRR], continuous=True)

    def test_unsupported_filetypes(self):
        with pytest.raises(NotImplementedError):
            mda.Universe(PSF, [DCD, DCD], continuous=True)
        # see issue 2353. The PDB reader has multiple format endings. To ensure 
        # the not implemented error is thrown  we  do a check here. A more  
        # careful test in the future would be a dummy reader with multiple 
        # formats, just in case PDB will allow continuous reading in the future.
        with pytest.raises(ValueError):
            mda.Universe(PDB, [PDB, XTC], continuous=True)


@pytest.mark.parametrize('l, ref', ([((0, 3), (3, 3), (4, 7)), (0, 1, 2)],
                                    [((0, 9), (0, 4)), (0, 1)],
                                    [((0, 3), (2, 2), (3, 3), (2, 6), (5, 9)), (0, 1, 3, 2, 4)],
                                    [((0, 2), (4, 9), (0, 4), (7, 9)), (0, 2, 1, 3)],
                                    [((0, 5), (2, 4), (2, 4), (2, 4), (3, 7)), (0, 1, 2, 3, 4)]
                                    ))
def test_multilevel_arg_sort(l, ref):
    indices = mda.coordinates.chain.multi_level_argsort(l)
    assert_equal(indices, ref)


@pytest.mark.parametrize('l, ref', ([((0, 4), (3, 6), (6, 9)), (0, 1, 2)],
                                    [((0, 3), (3, 4), (4, 7)), (0, 1, 2)],
                                    [((0, 3), (3, 5), (4, 7)), (0, 1, 2)],
                                    [((0, 3), (0, 4), (3, 5), (4, 7)), (1, 2, 3)],
                                    [((0, 3), (0, 4)), (1,)],
                                    [((0, 3), (0, 3)), (1,)],
                                    [((1, 3), (0, 4)), (0,)],
                                    [((0, 3), ), (0, )],
                                    [((0, 3), (5, 9)), (0, 1)],
                                    [((0, 3), (0, 3), (5, 9)), (1, 2)],
                                    [((0, 3), (0, 3), (0, 3), (5, 9)), (2, 3)],
                                    [((0, 5), (2, 4), (2, 4), (2, 4), (3, 7)), (0, 3, 4)],
                                    [((0, 3), (2, 4), (2, 4), (2, 4), (3, 7)), (0, 3, 4)],
                                    [((0, 3), (2, 4), (4, 7), (4, 7), (4, 7), (6, 9)), (0, 1, 4, 5)],
                                    [((0, 6), (2, 5), (2, 5), (2, 5), (3, 8)), (0, 3, 4)],
                                    ))
def test_filter_times(l, ref):
    indices = mda.coordinates.chain.filter_times(l, dt=1)
    assert_equal(indices, ref)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787

import pytest
import numpy as np
from numpy.testing import assert_almost_equal, assert_equal

from MDAnalysis.lib._augment import augment_coordinates, undo_augment
from MDAnalysis.lib.distances import apply_PBC, transform_StoR

# Find images for several query points,
# here in fractional coordinates
# Every element of qres tuple is (query, images)
qres = (
       ([0.1, 0.5, 0.5], [[1.1, 0.5, 0.5]]),     # box face
       ([0.5, 0.5, 0.5], []),                    # box center
       ([0.5, -0.1, 0.5], [[0.5, -0.1, 0.5]]),   # box face
       ([0.1, 0.1, 0.5], [[1.1, 0.1, 0.5],
                          [0.1, 1.1, 0.5],
                          [1.1, 1.1, 0.5]]),     # box edge
       ([0.5, -0.1, 1.1], [[0.5, -0.1, 0.1],
                           [0.5, 0.9, 1.1],
                           [0.5, -0.1, 1.1]]),   # box edge
       ([0.1, 0.1, 0.1], [[1.1, 0.1, 0.1],
                          [0.1, 1.1, 0.1],
                          [0.1, 0.1, 1.1],
                          [0.1, 1.1, 1.1],
                          [1.1, 1.1, 0.1],
                          [1.1, 0.1, 1.1],
                          [1.1, 1.1, 1.1]]),     # box vertex
       ([0.1, -0.1, 1.1], [[1.1, 0.9, 0.1],
                           [0.1, -0.1, 0.1],
                           [0.1, 0.9, 1.1],
                           [0.1, -0.1, 1.1],
                           [1.1, -0.1, 0.1],
                           [1.1, 0.9, 1.1],
                           [1.1, -0.1, 1.1]]),   # box vertex
       ([2.1, -3.1, 0.1], [[1.1, 0.9, 0.1],
                           [0.1, -0.1, 0.1],
                           [0.1, 0.9, 1.1],
                           [0.1, -0.1, 1.1],
                           [1.1, -0.1, 0.1],
                           [1.1, 0.9, 1.1],
                           [1.1, -0.1, 1.1]]),   # box vertex
       ([[0.1, 0.5, 0.5],
         [0.5, -0.1, 0.5]], [[1.1, 0.5, 0.5],
                             [0.5, -0.1, 0.5]])  # multiple queries
       )


@pytest.mark.parametrize('b', (
                         np.array([10, 10, 10, 90, 90, 90], dtype=np.float32),
                         np.array([10, 10, 10, 45, 60, 90], dtype=np.float32)
                         ))
@pytest.mark.parametrize('q, res', qres)
def test_augment(b, q, res):
    radius = 1.5
    q = transform_StoR(np.array(q, dtype=np.float32), b)
    if q.shape == (3, ):
        q = q.reshape((1, 3))
    q = apply_PBC(q, b)
    aug, mapping = augment_coordinates(q, b, radius)
    if aug.size > 0:
        aug = np.sort(aug, axis=0)
    else:
        aug = list()
    if len(res) > 0:
        cs = transform_StoR(np.array(res, dtype=np.float32), b)
        cs = np.sort(cs, axis=0)
    else:
        cs = list()
    assert_almost_equal(aug, cs, decimal=5)


@pytest.mark.parametrize('b', (
                         np.array([10, 10, 10, 90, 90, 90], dtype=np.float32),
                         np.array([10, 10, 10, 45, 60, 90], dtype=np.float32)
                         ))
@pytest.mark.parametrize('qres', qres)
def test_undoaugment(b, qres):
    radius = 1.5
    q = transform_StoR(np.array(qres[0], dtype=np.float32), b)
    if q.shape == (3, ):
        q = q.reshape((1, 3))
    q = apply_PBC(q, b)
    aug, mapping = augment_coordinates(q, b, radius)
    for idx, val in enumerate(aug):
        imageid = np.asarray([len(q) + idx], dtype=np.intp)
        assert_equal(mapping[idx], undo_augment(imageid, mapping, len(q))[0])



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
from unittest.mock import patch

import errno
import numpy as np
import os
import shutil
import subprocess

from numpy.testing import (assert_equal, assert_almost_equal)

from MDAnalysisTests import make_Universe
from MDAnalysisTests.datafiles import (
    PDB_sub_dry, PDB_sub_sol, TRR_sub_sol, TRR, XTC, GRO, PDB, CRD, PRMncdf,
    NCDF, XTC_sub_sol, COORDINATES_XTC, COORDINATES_TOPOLOGY, COORDINATES_TRR)
from MDAnalysisTests.coordinates.base import (MultiframeReaderTest,
                                              BaseReference, BaseWriterTest,
                                              assert_timestep_almost_equal)

import MDAnalysis as mda
from MDAnalysis.coordinates.base import Timestep
from MDAnalysis.coordinates import XDR


class _XDRReader_Sub(object):
    @pytest.fixture()
    def atoms(self):
        usol = mda.Universe(PDB_sub_sol, self.XDR_SUB_SOL)
        return usol.select_atoms("not resname SOL")

    def test_load_new_raises_ValueError(self):
        # should fail if we load universe with a trajectory with different
        # number of atoms when NOT using sub, same as before.
        udry = mda.Universe(PDB_sub_dry)
        with pytest.raises(ValueError):
            udry.load_new(self.XDR_SUB_SOL)

    def test_sub_coordinates(self, atoms):
        """
        load solvated trajectory into universe with unsolvated protein.
        """
        udry = mda.Universe(PDB_sub_dry)
        udry.load_new(self.XDR_SUB_SOL, sub=atoms.indices)
        ts = udry.atoms.ts
        assert_timestep_almost_equal(ts, atoms.ts)


class TestTRRReader_Sub(_XDRReader_Sub):
    XDR_SUB_SOL = TRR_sub_sol


class TestXTCReader_Sub(_XDRReader_Sub):
    XDR_SUB_SOL = XTC_sub_sol


class _GromacsReader(object):
    # This base class assumes same lengths and dt for XTC and TRR test cases!
    filename = None
    ref_unitcell = np.array(
        [80.017, 80.017, 80.017, 60., 60., 90.], dtype=np.float32)
    # computed with Gromacs: 362.26999999999998 nm**3 * 1000 A**3/nm**3
    ref_volume = 362270.0

    prec = 3

    @pytest.fixture(scope='class')
    def universe(self):
        return mda.Universe(GRO, self.filename, convert_units=True)

    def test_rewind_xdrtrj(self, universe):
        universe.trajectory.rewind()
        assert_equal(universe.coord.frame, 0, "rewinding to frame 1")
        assert universe.trajectory._xdr._has_offsets == 1

    def test_next_xdrtrj(self, universe):
        universe.trajectory.rewind()
        universe.trajectory.next()
        assert_equal(universe.coord.frame, 1, "loading frame 1")

    def test_jump_xdrtrj(self, universe):
        universe.trajectory[4]  # index is 0-based and frames are 0-based
        assert_equal(universe.coord.frame, 4, "jumping to frame 4")

    def test_jump_lastframe_xdrtrj(self, universe):
        universe.trajectory[-1]
        assert_equal(universe.coord.frame, 9,
                     "indexing last frame with trajectory[-1]")

    def test_slice_xdrtrj(self, universe):
        frames = [ts.frame for ts in universe.trajectory[2:9:3]]
        assert_equal(frames, [2, 5, 8], "slicing xdrtrj [2:9:3]")

    def test_reverse_xdrtrj(self, universe):
        frames = [ts.frame for ts in universe.trajectory[::-1]]
        assert_equal(frames, list(range(9, -1, -1)), "slicing xdrtrj [::-1]")

    def test_coordinates(self, universe):
        ca_nm = np.array(
            [[6.043369675, 7.385184479, 1.381425762]], dtype=np.float32)
        # coordinates in the base unit (needed for True)
        ca_Angstrom = ca_nm * 10.0
        universe.trajectory.rewind()
        universe.trajectory.next()
        universe.trajectory.next()
        assert_equal(universe.coord.frame, 2, "failed to step to frame 3")
        ca = universe.select_atoms('name CA and resid 122')
        # low precision match (2 decimals in A, 3 in nm) because the above are
        # the trr coords
        assert_almost_equal(
            ca.positions,
            ca_Angstrom,
            2,
            err_msg="coords of Ca of resid 122 do not "
            "match for frame 3")

    def test_unitcell(self, universe):
        """Test that xtc/trr unitcell is read correctly (Issue 34)"""
        universe.trajectory.rewind()
        uc = universe.coord.dimensions
        assert_almost_equal(
            uc,
            self.ref_unitcell,
            self.prec,
            err_msg="unit cell dimensions (rhombic dodecahedron)")

    def test_volume(self, universe):
        # need to reduce precision for test (nm**3 <--> A**3)
        universe.trajectory.rewind()
        vol = universe.coord.volume
        assert_almost_equal(
            vol,
            self.ref_volume,
            0,
            err_msg="unit cell volume (rhombic dodecahedron)")

    def test_dt(self, universe):
        assert_almost_equal(
            universe.trajectory.dt, 100.0, 4, err_msg="wrong timestep dt")

    def test_totaltime(self, universe):
        # test_totaltime(): need to reduce precision because dt is only precise
        # to ~4 decimals and accumulating the inaccuracy leads to even lower
        # precision in the totaltime (consequence of fixing Issue 64)
        assert_almost_equal(
            universe.trajectory.totaltime,
            900.0,
            3,
            err_msg="wrong total length of trajectory")

    def test_frame(self, universe):
        universe.trajectory[4]  # index is 0-based and frames are 0-based
        assert_equal(universe.trajectory.frame, 4, "wrong frame number")

    def test_time(self, universe):
        universe.trajectory[4]
        assert_almost_equal(
            universe.trajectory.time, 400.0, 3, err_msg="wrong time of frame")

    def test_get_Writer(self, universe, tmpdir):
        ext = os.path.splitext(self.filename)[1]
        outfile = str(tmpdir.join('xdr-reader-test' + ext))
        with universe.trajectory.Writer(outfile) as W:
            assert_equal(universe.trajectory.format, W.format)
            assert_equal(universe.atoms.n_atoms, W.n_atoms)

    def test_Writer(self, tmpdir):
        universe = mda.Universe(GRO, self.filename, convert_units=True)
        ext = os.path.splitext(self.filename)[1]
        outfile = str(tmpdir.join('/xdr-reader-test' + ext))
        with universe.trajectory.Writer(outfile) as W:
            W.write(universe.atoms)
            universe.trajectory.next()
            W.write(universe.atoms)

        universe.trajectory.rewind()
        u = mda.Universe(GRO, outfile)
        assert_equal(u.trajectory.n_frames, 2)
        # prec = 6: TRR test fails; here I am generous and take self.prec =
        # 3...
        assert_almost_equal(u.atoms.positions, universe.atoms.positions,
                            self.prec)

    def test_EOFraisesStopIteration(self, universe):
        def go_beyond_EOF():
            universe.trajectory[-1]
            universe.trajectory.next()

        with pytest.raises(StopIteration):
            go_beyond_EOF()


class TestXTCReader(_GromacsReader):
    filename = XTC


class TestXTCReaderClass(object):
    def test_with_statement(self):
        from MDAnalysis.coordinates.XTC import XTCReader

        try:
            with XTCReader(XTC) as trj:
                N = trj.n_frames
                frames = [ts.frame for ts in trj]
        except:
            raise AssertionError("with_statement not working for XTCReader")
        assert_equal(
            N,
            10,
            err_msg="with_statement: XTCReader reads wrong number of frames")
        assert_equal(
            frames,
            np.arange(0, N),
            err_msg="with_statement: XTCReader does not read all frames")


class TestTRRReader(_GromacsReader):
    filename = TRR

    def test_velocities(self, universe):
        # frame 0, v in nm/ps
        # from gmxdump -f MDAnalysisTests/data/adk_oplsaa.trr
        #      v[47675]={-7.86469e-01,  1.57479e+00,  2.79722e-01}
        #      v[47676]={ 2.70593e-08,  1.08052e-06,  6.97028e-07}
        v_native = np.array(
            [[-7.86469e-01, 1.57479e+00, 2.79722e-01],
             [2.70593e-08, 1.08052e-06, 6.97028e-07]],
            dtype=np.float32)

        # velocities in the MDA base unit A/ps (needed for True)
        v_base = v_native * 10.0
        universe.trajectory.rewind()
        assert_equal(universe.coord.frame, 0, "failed to read frame 1")

        assert_almost_equal(
            universe.trajectory.ts._velocities[[47675, 47676]],
            v_base,
            self.prec,
            err_msg="ts._velocities for indices 47675,47676 do not "
            "match known values")

        assert_almost_equal(
            universe.atoms.velocities[[47675, 47676]],
            v_base,
            self.prec,
            err_msg="velocities for indices 47675,47676 do not "
            "match known values")

        for index, v_known in zip([47675, 47676], v_base):
            assert_almost_equal(
                universe.atoms[index].velocity,
                v_known,
                self.prec,
                err_msg="atom[{0:d}].velocity does not match known values".
                format(index))


class _XDRNoConversion(object):
    filename = None

    @pytest.fixture()
    def universe(self):
        return mda.Universe(PDB, self.filename, convert_units=False)

    def test_coordinates(self, universe):
        # note: these are the native coordinates in nm
        ca_nm = np.array(
            [[6.043369675, 7.385184479, 1.381425762]], dtype=np.float32)
        universe.trajectory.rewind()
        universe.trajectory.next()
        universe.trajectory.next()
        assert_equal(universe.trajectory.ts.frame, 2,
                     "failed to step to frame 3")
        ca = universe.select_atoms('name CA and resid 122')
        # low precision match because we also look at the trr: only 3 decimals
        # in nm in xtc!
        assert_almost_equal(
            ca.positions,
            ca_nm,
            3,
            err_msg="native coords of Ca of resid 122 "
            "do not match for frame 3 with "
            "convert_units=False")


class TestXTCNoConversion(_XDRNoConversion):
    filename = XTC


class TestTRRNoConversion(_XDRNoConversion):
    filename = TRR


class _GromacsWriter(object):
    infilename = None  # XTC or TRR
    Writers = {
        '.trr': mda.coordinates.TRR.TRRWriter,
        '.xtc': mda.coordinates.XTC.XTCWriter,
    }

    @pytest.fixture(scope='class')
    def universe(self):
        return mda.Universe(GRO, self.infilename)

    @pytest.fixture()
    def Writer(self):
        ext = os.path.splitext(self.infilename)[1]
        return self.Writers[ext]

    @pytest.fixture()
    def outfile(self, tmpdir):
        ext = os.path.splitext(self.infilename)[1]
        return str(tmpdir.join('xdr-writer-test' + ext))

    def test_write_trajectory(self, universe, Writer, outfile):
        """Test writing Gromacs trajectories (Issue 38)"""
        with Writer(outfile, universe.atoms.n_atoms, dt=universe.trajectory.dt) as W:
            for ts in universe.trajectory:
                W.write(universe)

        uw = mda.Universe(GRO, outfile)

        # check that the coordinates are identical for each time step
        for orig_ts, written_ts in zip(universe.trajectory, uw.trajectory):
            assert_almost_equal(
                written_ts._pos,
                orig_ts._pos,
                3,
                err_msg="coordinate mismatch between "
                "original and written trajectory at "
                "frame %d (orig) vs %d (written)" % (orig_ts.frame,
                                                     written_ts.frame))

    def test_timestep_not_modified_by_writer(self, universe, Writer, outfile):
        trj = universe.trajectory
        ts = trj.ts

        trj[-1]  # last timestep (so that time != 0)
        x = ts._pos.copy()
        time = ts.time

        with Writer(outfile, trj.n_atoms, dt=trj.dt) as W:
            # last timestep (so that time != 0) (say it again, just in case...)
            trj[-1]
            W.write(universe)

        assert_equal(
            ts._pos,
            x,
            err_msg="Positions in Timestep were modified by writer.")
        assert_equal(
            ts.time, time, err_msg="Time in Timestep was modified by writer.")


class TestXTCWriter(_GromacsWriter):
    __test__ = True
    infilename = XTC


class TestTRRWriter(_GromacsWriter):
    __test__ = True
    infilename = TRR

    def test_velocities(self, universe, Writer, outfile):
        with Writer(outfile, universe.atoms.n_atoms, dt=universe.trajectory.dt) as W:
            for ts in universe.trajectory:
                W.write(universe)

        uw = mda.Universe(GRO, outfile)

        # check that the velocities are identical for each time step
        for orig_ts, written_ts in zip(universe.trajectory, uw.trajectory):
            assert_almost_equal(
                written_ts._velocities,
                orig_ts._velocities,
                3,
                err_msg="velocities mismatch between "
                "original and written trajectory at "
                "frame %d (orig) vs %d (written)" % (orig_ts.frame,
                                                     written_ts.frame))

    def test_gaps(self, universe, Writer, outfile):
        """Tests the writing and reading back of TRRs with gaps in any of
        the coordinates/velocities properties."""
        with Writer(outfile, universe.atoms.n_atoms, dt=universe.trajectory.dt) as W:
            for ts in universe.trajectory:
                # Inset some gaps in the properties: coords every 4 steps, vels
                # every 2.
                if ts.frame % 4 == 0:
                    ts.has_positions = False
                if ts.frame % 2 == 0:
                    ts.has_velocities = False
                W.write(universe)

        uw = mda.Universe(GRO, outfile)
        # check that the velocities are identical for each time step, except
        # for the gaps (that we must make sure to raise exceptions on).
        for orig_ts, written_ts in zip(universe.trajectory, uw.trajectory):
            if ts.frame % 4 != 0:
                assert_almost_equal(
                    written_ts.positions,
                    orig_ts.positions,
                    3,
                    err_msg="coordinates mismatch "
                    "between original and written "
                    "trajectory at frame {} (orig) "
                    "vs {} (written)".format(orig_ts.frame, written_ts.frame))
            else:
                with pytest.raises(mda.NoDataError):
                    getattr(written_ts, 'positions')

            if ts.frame % 2 != 0:
                assert_almost_equal(
                    written_ts.velocities,
                    orig_ts.velocities,
                    3,
                    err_msg="velocities mismatch "
                    "between original and written "
                    "trajectory at frame {} (orig) "
                    "vs {} (written)".format(orig_ts.frame, written_ts.frame))
            else:
                with pytest.raises(mda.NoDataError):
                    getattr(written_ts, 'velocities')


class _GromacsWriterIssue101(object):
    Writers = {
        '.trr': mda.coordinates.TRR.TRRWriter,
        '.xtc': mda.coordinates.XTC.XTCWriter,
    }
    ext = None  # set to '.xtc' or '.trr'
    prec = 3

    @pytest.fixture()
    def Writer(self):
        return self.Writers[self.ext]

    @pytest.fixture()
    def outfile(self, tmpdir):
        return str(tmpdir.join('/xdr-writer-issue101' + self.ext))

    def test_single_frame_GRO(self, Writer, outfile):
        self._single_frame(GRO, Writer, outfile)

    def test_single_frame_PDB(self, Writer, outfile):
        self._single_frame(PDB, Writer, outfile)

    def test_single_frame_CRD(self, Writer, outfile):
        self._single_frame(CRD, Writer, outfile)

    def _single_frame(self, filename, Writer, outfile):
        u = mda.Universe(filename)
        with Writer(outfile, u.atoms.n_atoms) as W:
            W.write(u.atoms)
        w = mda.Universe(filename, outfile)
        assert_equal(w.trajectory.n_frames, 1,
                     "single frame trajectory has wrong number of frames")
        assert_almost_equal(
            w.atoms.positions,
            u.atoms.positions,
            self.prec,
            err_msg="coordinates do not match for {0!r}".format(filename))


class TestXTCWriterSingleFrame(_GromacsWriterIssue101):
    ext = ".xtc"
    prec = 2


class TestTRRWriterSingleFrame(_GromacsWriterIssue101):
    ext = ".trr"


class _GromacsWriterIssue117(object):
    """Issue 117: Cannot write XTC or TRR from AMBER NCDF"""
    ext = None
    prec = 5

    @pytest.fixture()
    def universe(self):
        return mda.Universe(PRMncdf, NCDF)

    def test_write_trajectory(self, universe, tmpdir):
        """Test writing Gromacs trajectories from AMBER NCDF (Issue 117)"""
        outfile = str(tmpdir.join('xdr-writer-issue117' + self.ext))
        with mda.Writer(outfile, n_atoms=universe.atoms.n_atoms) as W:
            for ts in universe.trajectory:
                W.write(universe)

        uw = mda.Universe(PRMncdf, outfile)

        # check that the coordinates are identical for each time step
        for orig_ts, written_ts in zip(universe.trajectory, uw.trajectory):
            assert_almost_equal(
                written_ts._pos,
                orig_ts._pos,
                self.prec,
                err_msg="coordinate mismatch "
                "between original and written "
                "trajectory at frame %d (orig) vs %d "
                "(written)" % (orig_ts.frame, written_ts.frame))


class TestXTCWriterIssue117(_GromacsWriterIssue117):
    __test__ = True
    ext = ".xtc"
    prec = 2


class TestTRRWriterIssue117(_GromacsWriterIssue117):
    __test__ = True
    ext = ".trr"


def test_triclinic_box():
    """Test coordinates.core.triclinic_box() (Issue 61)"""
    unitcell = np.array([80.017, 55, 100.11, 60.00, 30.50, 90.00])
    box = mda.coordinates.core.triclinic_vectors(unitcell)
    new_unitcell = mda.coordinates.core.triclinic_box(box[0], box[1], box[2])
    assert_almost_equal(
        new_unitcell,
        unitcell,
        3,
        err_msg="unitcell round-trip connversion failed (Issue 61)")


class XTCReference(BaseReference):
    def __init__(self):
        super(XTCReference, self).__init__()
        self.trajectory = COORDINATES_XTC
        self.topology = COORDINATES_TOPOLOGY
        self.reader = mda.coordinates.XTC.XTCReader
        self.writer = mda.coordinates.XTC.XTCWriter
        self.ext = 'xtc'
        self.prec = 3
        self.changing_dimensions = True


class TestXTCReader_2(MultiframeReaderTest):
    @staticmethod
    @pytest.fixture()
    def ref():
        return XTCReference()


class TestXTCWriter_2(BaseWriterTest):
    @staticmethod
    @pytest.fixture()
    def ref():
        return XTCReference()

    def test_different_precision(self, ref, tmpdir):
        out = 'precision-test' + ref.ext
        # store more then 9 atoms to enable compression
        n_atoms = 40
        with tmpdir.as_cwd():
            with ref.writer(out, n_atoms, precision=5) as w:
                u = make_Universe(size=(n_atoms, 1, 1), trajectory=True)
                u.trajectory.ts.positions = np.random.random(size=(n_atoms, 3))
                w.write(u)
            xtc = mda.lib.formats.libmdaxdr.XTCFile(out)
            frame = xtc.read()
            assert_equal(len(xtc), 1)
            assert_equal(xtc.n_atoms, n_atoms)
            assert_equal(frame.prec, 10.0**5)


class TRRReference(BaseReference):
    def __init__(self):
        super(TRRReference, self).__init__()
        self.trajectory = COORDINATES_TRR
        self.topology = COORDINATES_TOPOLOGY
        self.changing_dimensions = True
        self.reader = mda.coordinates.TRR.TRRReader
        self.writer = mda.coordinates.TRR.TRRWriter
        self.ext = 'trr'
        self.prec = 3
        self.first_frame.velocities = self.first_frame.positions / 10
        self.first_frame.forces = self.first_frame.positions / 100

        self.second_frame.velocities = self.second_frame.positions / 10
        self.second_frame.forces = self.second_frame.positions / 100

        self.last_frame.velocities = self.last_frame.positions / 10
        self.last_frame.forces = self.last_frame.positions / 100

        self.jump_to_frame.velocities = self.jump_to_frame.positions / 10
        self.jump_to_frame.forces = self.jump_to_frame.positions / 100

    def iter_ts(self, i):
        ts = self.first_frame.copy()
        ts.positions = 2**i * self.first_frame.positions
        ts.velocities = ts.positions / 10
        ts.forces = ts.positions / 100
        ts.time = i
        ts.frame = i
        return ts


class TestTRRReader_2(MultiframeReaderTest):
    @staticmethod
    @pytest.fixture()
    def ref():
        return TRRReference()


class TestTRRWriter_2(BaseWriterTest):
    @staticmethod
    @pytest.fixture()
    def ref():
        return TRRReference()

    # tests writing and reading in one!
    def test_lambda(self, ref, universe, tmpdir):
        outfile = 'write-lambda-test' + ref.ext

        with tmpdir.as_cwd():
            with ref.writer(outfile, universe.trajectory.n_atoms) as W:
                for i, ts in enumerate(universe.trajectory):
                    ts.data['lambda'] = i / float(universe.trajectory.n_frames)
                    W.write(universe)

            reader = ref.reader(outfile)
            for i, ts in enumerate(reader):
                assert_almost_equal(ts.data['lambda'], i / float(reader.n_frames))


class _GromacsReader_offsets(object):

    # This base class assumes same lengths and dt for XTC and TRR test cases!
    filename = None
    ref_unitcell = np.array(
        [80.017, 80.017, 80.017, 60., 60., 90.], dtype=np.float32)
    # computed with Gromacs: 362.26999999999998 nm**3 * 1000 A**3/nm**3
    ref_volume = 362270.0
    ref_offsets = None
    _reader = None
    prec = 3

    @pytest.fixture(scope='class')
    def traj(self, tmpdir_factory):
        # copy of original test trajectory in a temporary folder. This is
        # needed since offsets are automatically generated in the same
        # directory. Here we also clean up nicely all files we generate
        tmpdir = tmpdir_factory.mktemp('xtc')
        shutil.copy(self.filename, str(tmpdir))
        traj = str(tmpdir.join(os.path.basename(self.filename)))
        # ensure initialization of offsets
        self._reader(traj)
        return traj

    @pytest.fixture()
    def trajectory(self, traj):
        return self._reader(traj)

    def test_offsets(self, trajectory, traj):
        trajectory._read_offsets(store=True)
        assert_almost_equal(
            trajectory._xdr.offsets,
            self.ref_offsets,
            err_msg="wrong frame offsets")

        outfile_offsets = XDR.offsets_filename(traj)
        saved_offsets = XDR.read_numpy_offsets(outfile_offsets)

        assert isinstance(saved_offsets, dict), \
            "read_numpy_offsets did not return a dict"
        assert_almost_equal(
            trajectory._xdr.offsets,
            saved_offsets['offsets'],
            err_msg="error saving frame offsets")
        assert_almost_equal(
            self.ref_offsets,
            saved_offsets['offsets'],
            err_msg="saved frame offsets don't match "
            "the known ones")

        trajectory._load_offsets()
        assert_almost_equal(
            trajectory._xdr.offsets,
            self.ref_offsets,
            err_msg="error loading frame offsets")
        assert_equal(saved_offsets['ctime'], os.path.getctime(traj))
        assert_equal(saved_offsets['size'], os.path.getsize(traj))

    def test_reload_offsets(self, traj):
        self._reader(traj, refresh_offsets=True)

    def test_nonexistant_offsets_file(self, traj):
        # assert that a nonexistant file returns False during read-in
        outfile_offsets = XDR.offsets_filename(traj)
        with patch.object(np, "load") as np_load_mock:
            np_load_mock.side_effect = IOError
            saved_offsets = XDR.read_numpy_offsets(outfile_offsets)
            assert_equal(saved_offsets, False)

    def test_reload_offsets_if_offsets_readin_fails(self, trajectory):
        # force the np.load call that is called in read_numpy_offsets
        # during _load_offsets to give an IOError
        # ensure that offsets are then read-in from the trajectory
        with patch.object(np, "load") as np_load_mock:
            np_load_mock.side_effect = IOError
            trajectory._load_offsets()
            assert_almost_equal(
                trajectory._xdr.offsets,
                self.ref_offsets,
                err_msg="error loading frame offsets")

    def test_persistent_offsets_size_mismatch(self, traj):
        # check that stored offsets are not loaded when trajectory
        # size differs from stored size
        fname = XDR.offsets_filename(traj)
        saved_offsets = XDR.read_numpy_offsets(fname)

        assert isinstance(saved_offsets, dict), \
            "read_numpy_offsets did not return a dict"

        saved_offsets['size'] += 1
        with open(fname, 'wb') as f:
            np.savez(f, **saved_offsets)

        with pytest.warns(UserWarning, match="Reload offsets"):
            self._reader(traj)

    def test_persistent_offsets_ctime_mismatch(self, traj):
        # check that stored offsets are not loaded when trajectory
        # ctime differs from stored ctime
        fname = XDR.offsets_filename(traj)
        saved_offsets = XDR.read_numpy_offsets(fname)

        assert isinstance(saved_offsets, dict), \
            "read_numpy_offsets did not return a dict"

        saved_offsets['ctime'] += 1
        with open(fname, 'wb') as f:
            np.savez(f, **saved_offsets)

        with pytest.warns(UserWarning, match="Reload offsets"):
            self._reader(traj)

    def test_persistent_offsets_natoms_mismatch(self, traj):
        # check that stored offsets are not loaded when trajectory
        # ctime differs from stored ctime
        fname = XDR.offsets_filename(traj)
        saved_offsets = XDR.read_numpy_offsets(fname)

        assert isinstance(saved_offsets, dict), \
            "read_numpy_offsets did not return a dict"

        saved_offsets['n_atoms'] += 1
        np.savez(fname, **saved_offsets)

        with pytest.warns(UserWarning, match="Reload offsets"):
            self._reader(traj)

    def test_persistent_offsets_last_frame_wrong(self, traj):
        fname = XDR.offsets_filename(traj)
        saved_offsets = XDR.read_numpy_offsets(fname)

        assert isinstance(saved_offsets, dict), \
            "read_numpy_offsets did not return a dict"

        idx_frame = 3
        saved_offsets['offsets'][idx_frame] += 42
        np.savez(fname, **saved_offsets)

        with pytest.warns(UserWarning, match="seek failed"):
            reader = self._reader(traj)
            reader[idx_frame]

    def test_unsupported_format(self, traj):
        fname = XDR.offsets_filename(traj)
        saved_offsets = XDR.read_numpy_offsets(fname)

        assert isinstance(saved_offsets, dict), \
            "read_numpy_offsets did not return a dict"

        idx_frame = 3
        saved_offsets.pop('n_atoms')
        np.savez(fname, **saved_offsets)

        # ok as long as this doesn't throw
        reader = self._reader(traj)
        reader[idx_frame]

    def test_persistent_offsets_readonly(self, tmpdir):
        shutil.copy(self.filename, str(tmpdir))

        if os.name == 'nt':
            # Windows platform has a unique way to deny write access
            subprocess.call("icacls {fname} /deny Users:W".format(fname=tmpdir),
                            shell=True)
        else:
            os.chmod(str(tmpdir), 0o555)

        filename = str(tmpdir.join(os.path.basename(self.filename)))
        # try to write a offsets file
        self._reader(filename)
        assert_equal(os.path.exists(XDR.offsets_filename(filename)), False)


class TestXTCReader_offsets(_GromacsReader_offsets):
    __test__ = True
    filename = XTC
    ref_offsets = np.array([
        0, 165188, 330364, 495520, 660708, 825872, 991044, 1156212, 1321384,
        1486544
    ])
    _reader = mda.coordinates.XTC.XTCReader


class TestTRRReader_offsets(_GromacsReader_offsets):
    __test__ = True
    filename = TRR
    ref_offsets = np.array([
        0, 1144464, 2288928, 3433392, 4577856, 5722320, 6866784, 8011248,
        9155712, 10300176
    ])
    _reader = mda.coordinates.TRR.TRRReader



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest

import MDAnalysis as mda

from MDAnalysisTests.topology.base import ParserBase
from MDAnalysisTests.datafiles import GSD
from MDAnalysisTests.datafiles import GSD_bonds
from numpy.testing import assert_equal
import os

class GSDBase(ParserBase):
    parser = mda.topology.GSDParser.GSDParser
    expected_attrs = ['ids', 'names', 'resids', 'resnames', 'masses',
                      'charges', 'radii',
                      'bonds', 'angles', 'dihedrals', 'impropers']
    expected_n_bonds = 0
    expected_n_angles = 0
    expected_n_dihedrals = 0
    expected_n_impropers = 0
    
    def test_attr_size(self, top):
        assert len(top.ids) == top.n_atoms
        assert len(top.names) == top.n_atoms
        assert len(top.resids) == top.n_residues
        assert len(top.resnames) == top.n_residues
    
    def test_atoms(self, top):
        assert top.n_atoms == self.expected_n_atoms

    def test_bonds(self, top):
        assert len(top.bonds.values) == self.expected_n_bonds
        if self.expected_n_bonds:
            assert isinstance(top.bonds.values[0], tuple)
        else:
            assert top.bonds.values == []

    def test_angles(self, top):
        assert len(top.angles.values) == self.expected_n_angles
        if self.expected_n_angles:
            assert isinstance(top.angles.values[0], tuple)
        else:
            assert top.angles.values == []

    def test_dihedrals(self, top):
        assert len(top.dihedrals.values) == self.expected_n_dihedrals
        if self.expected_n_dihedrals:
            assert isinstance(top.angles.values[0], tuple)
        else:
            assert top.dihedrals.values == []
    
    def test_impropers(self, top):
        assert len(top.impropers.values) == self.expected_n_impropers
        if self.expected_n_impropers:
            assert isinstance(top.angles.values[0], tuple)
        else:
            assert top.impropers.values == []


class TestGSDParser(GSDBase):
    ref_filename = GSD
    expected_n_atoms = 5832
    expected_n_residues = 648
    expected_n_segments = 1

    

class TestGSDParserBonds(GSDBase):
    ref_filename = GSD_bonds
    expected_n_atoms = 490
    expected_n_bonds = 441
    expected_n_angles = 392
    expected_n_dihedrals = 343
    expected_n_residues = 1
    expected_n_segments = 1

    def test_bonds_identity(self, top):
        vals = top.bonds.values
        for b in ((0, 1), (1, 2), (2, 3), (3, 4)):
            assert (b in vals) or (b[::-1] in vals)
        assert ((0, 450) not in vals)

    def test_angles_identity(self, top):
        vals = top.angles.values
        for b in ((0, 1, 2), (1, 2, 3), (2, 3, 4), (3, 4, 5)):
            assert (b in vals) or (b[::-1] in vals)
        assert ((0, 350, 450) not in vals)

    def test_dihedrals_identity(self, top):
        vals = top.dihedrals.values
        for b in ((0, 1, 2, 3), (1, 2, 3, 4), (2, 3, 4, 5), (3, 4, 5, 6)):
            assert (b in vals) or (b[::-1] in vals)
        assert ((0, 250, 350, 450) not in vals)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
import pytest
from numpy.testing import (
    assert_almost_equal,
    assert_equal,
)

from MDAnalysisTests import make_Universe
from MDAnalysisTests.datafiles import (
    COORDINATES_XYZ, COORDINATES_TRR,
    GRO, TRR,
    GRO_velocity, PDB_xvf, TRR_xvf
)

import MDAnalysis
from MDAnalysis import NoDataError


def assert_not_view(arr):
    assert arr.flags['OWNDATA'] is True


def assert_correct_errormessage(func, var):
    errmsg = "Timestep has no {}".format(var)
    try:
        func[0](*func[1:])
    except NoDataError as e:
        assert errmsg in e.args[0]
    else:
        pytest.fail()


@pytest.mark.parametrize('pos,vel,force', (
    (True, False, False),
    (True, True, False),
    (True, False, True),
    (True, True, True),
), indirect=True)
class TestAtomGroupTrajAccess(object):
    """
    For AtomGroup and Atom access:

    if present:
      - check return type
      - check dtype of array
      - check not a view of original (should always be copy!)
      - check the actual values returned
    if not present in trajectory:
      - check we get a NoDataError
      - check the error message of NDE

    For AtomGroup and Atom setting:

    if present:
      - check AtomGroup value is updated
      - check value in master Timestep object is updated
    if not present, check we get proper NoDataError on setting
    """
    @pytest.fixture()
    def pos(self, request):
        return request.param

    @pytest.fixture()
    def force(self, request):
        return request.param

    @pytest.fixture()
    def vel(self, request):
        return request.param

    @pytest.fixture()
    def u(self, pos, vel, force):
        return make_Universe(trajectory=pos, velocities=vel, forces=force)

    def test_atomgroup_positions_access(self, u):
        ag = u.atoms[10:20]
        ag_pos = ag.positions

        assert isinstance(ag_pos, np.ndarray)
        assert ag_pos.dtype == np.float32
        assert_not_view(ag_pos)
        assert_equal(ag_pos, u.trajectory.ts.positions[10:20])

    def test_atomgroup_velocities_access(self, u, vel):
        ag = u.atoms[10:20]

        if vel:
            ag_vel = ag.velocities

            assert isinstance(ag_vel, np.ndarray)
            assert ag_vel.dtype == np.float32
            assert_not_view(ag_vel)
            assert_equal(ag_vel, u.trajectory.ts.velocities[10:20])
        else:
            with pytest.raises(NoDataError):
                getattr(ag, 'velocities')
            assert_correct_errormessage((getattr, ag, 'velocities'),
                                        'velocities')

    def test_atomgroup_forces_access(self, u, force):
        ag = u.atoms[10:20]

        if force:
            ag_for = ag.forces

            assert isinstance(ag_for, np.ndarray)
            assert ag_for.dtype == np.float32
            assert_not_view(ag_for)
            assert_equal(ag_for, u.trajectory.ts.forces[10:20])
        else:
            with pytest.raises(NoDataError):
                getattr(ag, 'forces')
            assert_correct_errormessage((getattr, ag, 'forces'), 'forces')

    def test_atom_position_access(self, u):
        at = u.atoms[55]

        at_pos = at.position

        assert isinstance(at_pos, np.ndarray)
        assert at_pos.dtype == np.float32
        assert_not_view(at_pos)
        assert_equal(at_pos, u.trajectory.ts.positions[55])

    def test_atom_velocity_access(self, u, vel):
        at = u.atoms[55]

        if vel:
            at_vel = at.velocity

            assert isinstance(at_vel, np.ndarray)
            assert at_vel.dtype == np.float32
            assert_not_view(at_vel)
            assert_equal(at_vel, u.trajectory.ts.velocities[55])
        else:
            with pytest.raises(NoDataError):
                getattr(at, 'velocity')
            assert_correct_errormessage(
                (getattr, at, 'velocity'), 'velocities')

    def test_atom_force_access(self, u, force):
        at = u.atoms[55]

        if force:
            at_for = at.force

            assert isinstance(at_for, np.ndarray)
            assert at_for.dtype == np.float32
            assert_not_view(at_for)
            assert_equal(at_for, u.trajectory.ts.forces[55])
        else:
            with pytest.raises(NoDataError):
                getattr(at, 'force')
            assert_correct_errormessage((getattr, at, 'force'), 'forces')

    def test_atomgroup_positions_setting(self, u):
        ag = u.atoms[[101, 107, 109]]

        new = np.array([[72.4, 64.5, 74.7],
                        [124.6, 15.6, -1.11],
                        [25.2, -66.6, 0]])

        ag.positions = new

        assert_almost_equal(ag.positions, new, decimal=5)
        assert_almost_equal(u.trajectory.ts.positions[[101, 107, 109]],
                            new, decimal=5)

    def test_atomgroup_velocities_setting(self, u, vel):
        ag = u.atoms[[101, 107, 109]]

        new = np.array([[72.4, 64.5, 74.7],
                        [124.6, 15.6, -1.11],
                        [25.2, -66.6, 0]]) + 0.1

        if vel:
            ag.velocities = new

            assert_almost_equal(ag.velocities, new, decimal=5)
            assert_almost_equal(
                u.trajectory.ts.velocities[[101, 107, 109]], new, decimal=5)
        else:
            with pytest.raises(NoDataError):
                setattr(ag, 'velocities', new)
            assert_correct_errormessage((setattr, ag, 'velocities', new),
                                        'velocities')

    def test_atomgroup_forces_setting(self, u, force):
        ag = u.atoms[[101, 107, 109]]

        new = np.array([[72.4, 64.5, 74.7],
                        [124.6, 15.6, -1.11],
                        [25.2, -66.6, 0]]) + 0.2

        if force:
            ag.forces = new

            assert_almost_equal(ag.forces, new, decimal=5)
            assert_almost_equal(u.trajectory.ts.forces[[101, 107, 109]],
                                new, decimal=5)
        else:
            with pytest.raises(NoDataError):
                setattr(ag, 'forces', new)
            assert_correct_errormessage((setattr, ag, 'forces', new), 'forces')

    def test_atom_position_setting(self, u):
        at = u.atoms[94]

        new = np.array([58.3, -10.1, 0.001])

        at.position = new

        assert_almost_equal(at.position, new, decimal=5)
        assert_almost_equal(u.trajectory.ts.positions[94], new, decimal=5)

    def test_atom_velocity_setting(self, u, vel):
        at = u.atoms[94]

        new = np.array([58.3, -10.1, 0.001]) + 0.1

        if vel:
            at.velocity = new

            assert_almost_equal(at.velocity, new, decimal=5)
            assert_almost_equal(u.trajectory.ts.velocities[94], new,
                                decimal=5)
        else:
            with pytest.raises(NoDataError):
                setattr(at, 'velocity', new)
            assert_correct_errormessage((setattr, at, 'velocity', new),
                                        'velocities')

    def test_atom_force_setting(self, u, force):
        at = u.atoms[94]

        new = np.array([58.3, -10.1, 0.001]) + 0.2

        if force:
            at.force = new

            assert_almost_equal(at.force, new, decimal=5)
            assert_almost_equal(u.trajectory.ts.forces[94], new,
                                decimal=5)
        else:
            with pytest.raises(NoDataError):
                setattr(at, 'force', new)
            assert_correct_errormessage((setattr, at, 'force', new), 'forces')


class TestAtom_ForceVelocity(object):
    @pytest.fixture()
    def u(self):
        return MDAnalysis.Universe(PDB_xvf, TRR_xvf)

    @pytest.fixture()
    def a(self, u):
        return u.atoms[0]

    def test_atom_force_get(self, u, a):
        assert_equal(a.force, u.atoms.forces[0])

    def test_atom_velocity_get(self, u, a):
        assert_equal(a.velocity, u.atoms.velocities[0])

    def test_atom_force_set(self, u, a):
        ref = np.arange(3)
        a.force = ref

        assert_equal(a.force, ref)
        assert_equal(u.atoms.forces[0], ref)

    def test_atom_velocity_set(self, u, a):
        ref = np.arange(3)
        a.velocity = ref

        assert_equal(a.velocity, ref)
        assert_equal(u.atoms.velocities[0], ref)

    def test_pos_iteration(self, u, a):
        ag = u.atoms[[0]]

        val = np.array([a.position for ts in u.trajectory])
        ref = np.array([ag.positions[0] for ts in u.trajectory])

        assert_equal(val, ref)

    def test_vel_iteration(self, u, a):
        ag = u.atoms[[0]]

        val = np.array([a.velocity for ts in u.trajectory])
        ref = np.array([ag.velocities[0] for ts in u.trajectory])

        assert_equal(val, ref)

    def test_for_iteration(self, u, a):
        ag = u.atoms[[0]]

        val = np.array([a.force for ts in u.trajectory])
        ref = np.array([ag.forces[0] for ts in u.trajectory])

        assert_equal(val, ref)


class TestGROVelocities(object):
    prec = 3

    @pytest.fixture()
    def reference_velocities(self):
        return np.array(
            [[-101.227, -0.57999998, 0.43400002],
             [8.08500004, 3.19099998, -7.79099989],
             [-9.04500008, -26.46899986, 13.17999935],
             [2.51899981, 3.1400001, -1.73399997],
             [-10.64100075, -11.34899998, 0.257],
             [19.42700005, -8.21600056, -0.24399999]], dtype=np.float32)

    def testParse_velocities(self, reference_velocities):
        # read the velocities from the GRO_velocity file and compare the AtomGroup and individual Atom velocities
        # parsed with the reference values:
        u = MDAnalysis.Universe(GRO_velocity)
        all_atoms = u.select_atoms('all')
        # check for read-in and unit conversion for .gro file velocities for the entire AtomGroup:
        assert_almost_equal(all_atoms.velocities, reference_velocities,
                            self.prec,
                            err_msg="problem reading .gro file velocities")
        # likewise for each individual atom (to be robust--in case someone alters the individual atom property code):
        assert_almost_equal(all_atoms[0].velocity, reference_velocities[0],
                            self.prec,
                            err_msg="problem reading .gro file velocities")
        assert_almost_equal(all_atoms[1].velocity, reference_velocities[1],
                            self.prec,
                            err_msg="problem reading .gro file velocities")
        assert_almost_equal(all_atoms[2].velocity, reference_velocities[2],
                            self.prec,
                            err_msg="problem reading .gro file velocities")
        assert_almost_equal(all_atoms[3].velocity, reference_velocities[3],
                            self.prec,
                            err_msg="problem reading .gro file velocities")
        assert_almost_equal(all_atoms[4].velocity, reference_velocities[4],
                            self.prec,
                            err_msg="problem reading .gro file velocities")
        assert_almost_equal(all_atoms[5].velocity, reference_velocities[5],
                            self.prec,
                            err_msg="problem reading .gro file velocities")


class TestTRRForces(object):
    prec = 6

    @pytest.fixture()
    def universe(self):
        return MDAnalysis.Universe(PDB_xvf, TRR_xvf)

    @pytest.fixture()
    def reference_mean_protein_force(self):
        reference_mean_protein_force_native = np.array(
            [3.4609879271822823, -0.63302345167392804, -1.0587882545813336],
            dtype=np.float32)
        return reference_mean_protein_force_native / 10

    def testForces(self, universe, reference_mean_protein_force):
        protein = universe.select_atoms("protein")
        assert_equal(len(protein), 918)
        mean_F = np.mean(
            [protein.forces.mean(axis=0) for ts in universe.trajectory], axis=0)
        assert_almost_equal(mean_F, reference_mean_protein_force, self.prec,
                            err_msg="mean force on protein over whole trajectory does not match")


class TestTRRForcesNativeUnits(TestTRRForces):
    @pytest.fixture()
    def universe(self):
        return MDAnalysis.Universe(PDB_xvf, TRR_xvf, convert_units=False)

    @pytest.fixture()
    def reference_mean_protein_force(self):
        reference_mean_protein_force_native = np.array(
            [3.4609879271822823, -0.63302345167392804, -1.0587882545813336],
            dtype=np.float32)
        return reference_mean_protein_force_native


class TestAtomGroupVelocities(object):
    """Tests of velocity-related functions in AtomGroup"""

    @pytest.fixture()
    def universe(self):
        return MDAnalysis.Universe(GRO, TRR)

    @pytest.fixture()
    def ag(self, universe):
        return universe.select_atoms("bynum 12:42")

    def test_get_velocities(self, ag):
        v = ag.velocities
        assert np.any(np.abs(v) > 1e-6), "velocities should be non-zero"

    def test_velocities(self, universe):
        ag = universe.atoms[42:45]
        ref_v = np.array([
            [-3.61757946, -4.9867239, 2.46281552],
            [2.57792854, 3.25411797, -0.75065529],
            [13.91627216, 30.17778587, -12.16669178]])
        v = ag.velocities
        assert_almost_equal(v, ref_v,
                            err_msg="velocities were not read correctly")

    def test_set_velocities(self, ag):
        ag = ag
        v = ag.velocities - 2.7271
        ag.velocities = v
        assert_almost_equal(ag.velocities, v,
                            err_msg="messages were not set to new value")


class TestAtomGroupForces(object):
    """Tests of velocity-related functions in AtomGroup"""

    @pytest.fixture()
    def universe(self):
        return MDAnalysis.Universe(COORDINATES_XYZ, COORDINATES_TRR)

    @pytest.fixture()
    def ag(self, universe):
        return universe.atoms[1:4]

    def test_get_forces(self, ag):
        v = ag.forces
        assert np.any(np.abs(v) > 1e-6), "forces should be non-zero"

    def test_forces(self, universe):
        ag = universe.atoms[1:4]
        ref_v = np.arange(9).reshape(3, 3) * .01 + .03
        v = ag.forces
        assert_almost_equal(v, ref_v, err_msg="forces were not read correctly")

    def test_set_forces(self, ag):
        v = ag.forces - 2.7271
        ag.forces = v
        assert_almost_equal(ag.forces, v,
                            err_msg="messages were not set to new value")



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
import pickle

import MDAnalysis as mda
from numpy.testing import (
    TestCase,
    assert_equal
)

import gc

from MDAnalysisTests.datafiles import PDB_small


class TestAtomGroupPickle(object):
    """Set up hopefully unique universes."""

    # _n marks named universes/atomgroups/pickled strings

    @staticmethod
    @pytest.fixture()
    def universe():
        return mda.Universe(PDB_small, PDB_small, PDB_small)

    @staticmethod
    @pytest.fixture()
    def ag(universe):
        return universe.atoms[:20]

    @staticmethod
    @pytest.fixture()
    def ag_2(universe):
        return universe.atoms[10:20]

    @staticmethod
    @pytest.fixture()
    def pickle_str(ag):
        return pickle.dumps(ag, protocol=pickle.HIGHEST_PROTOCOL)

    @staticmethod
    @pytest.fixture()
    def pickle_str_two_ag(ag, ag_2):
        return pickle.dumps((ag, ag_2), protocol=pickle.HIGHEST_PROTOCOL)

    @staticmethod
    @pytest.fixture()
    def pickle_str_ag_with_universe_f(ag, universe):
        return pickle.dumps((universe, ag), protocol=pickle.HIGHEST_PROTOCOL)

    @staticmethod
    @pytest.fixture()
    def pickle_str_ag_with_universe(ag, universe):
        return pickle.dumps((ag, universe), protocol=pickle.HIGHEST_PROTOCOL)

    def test_unpickle(self, pickle_str, ag, universe):
        """Test that an AtomGroup can be unpickled (Issue 293)"""
        newag = pickle.loads(pickle_str)
        # Can unpickle
        assert_equal(ag.indices, newag.indices)

    def test_pickle_unpickle_empty(self, universe):
        """Test that an empty AtomGroup can be pickled/unpickled (Issue 293)"""
        ag = universe.atoms[[]]
        pickle_str = pickle.dumps(ag, protocol=pickle.HIGHEST_PROTOCOL)
        newag = pickle.loads(pickle_str)
        assert len(newag) == 0

    def test_unpickle_two_ag(self, pickle_str_two_ag):
        newag, newag2 = pickle.loads(pickle_str_two_ag)
        assert newag.universe is newag2.universe, (
            "Two AtomGroups are unpickled to two different Universes"
        )

    def test_unpickle_ag_with_universe_f(self,
                                         pickle_str_ag_with_universe_f):
        newu, newag = pickle.loads(pickle_str_ag_with_universe_f)
        assert newag.universe is newu, (
            "AtomGroup is not unpickled to the bound Universe"
            "when Universe is pickled first"
        )

    def test_unpickle_ag_with_universe(self,
                                       pickle_str_ag_with_universe):
        newag, newu = pickle.loads(pickle_str_ag_with_universe)
        assert newag.universe is newu, (
                "AtomGroup is not unpickled to the bound Universe"
                "when AtomGroup is pickled first"
        )


class TestPicklingUpdatingAtomGroups(object):

    @staticmethod
    @pytest.fixture()
    def u():
        return mda.Universe(PDB_small)

    def test_pickling_uag(self, u):
        ag = u.atoms[:100]
        uag = ag.select_atoms('name C', updating=True)
        pickle_str = pickle.dumps(uag, protocol=pickle.HIGHEST_PROTOCOL)
        new_uag = pickle.loads(pickle_str)

        assert_equal(uag.indices, new_uag.indices)

    def test_pickling_uag_of_uag(self, u):
        uag1 = u.select_atoms('name C or name H', updating=True)
        uag2 = uag1.select_atoms('name C', updating=True)
        pickle_str = pickle.dumps(uag2, protocol=pickle.HIGHEST_PROTOCOL)
        new_uag2 = pickle.loads(pickle_str)

        assert_equal(uag2.indices, new_uag2.indices)



from io import StringIO
from collections import defaultdict

from numpy.testing import (
    assert_equal, assert_array_equal,)
import pytest

import MDAnalysis
from MDAnalysis.analysis.hydrogenbonds.wbridge_analysis import (
    WaterBridgeAnalysis, )


class TestWaterBridgeAnalysis(object):
    @staticmethod
    @pytest.fixture(scope='class')
    def universe_empty():
        '''A universe with no hydrogen bonds'''
        grofile = '''Test gro file
5
    1ALA      N    1   0.000   0.000   0.000
    1ALA      H    2   0.100   0.000   0.000
    2SOL     OW    3   3.000   0.000   0.000
    4ALA      H    4   0.500   0.000   0.000
    4ALA      N    5   0.600   0.000   0.000
  1.0   1.0   1.0'''
        u = MDAnalysis.Universe(StringIO(grofile), format='gro')
        return u

    @staticmethod
    @pytest.fixture(scope='class')
    def universe_DA():
        '''A universe with one hydrogen bond acceptor bonding to a hydrogen bond
        donor'''
        grofile = '''Test gro file
3
    1ALA      N    1   0.000   0.000   0.000
    1ALA      H    2   0.100   0.000   0.000
    4ALA      O    3   0.300   0.000   0.000
  1.0   1.0   1.0'''
        u = MDAnalysis.Universe(StringIO(grofile), format='gro')
        return u

    @staticmethod
    @pytest.fixture(scope='class')
    def universe_DA_PBC():
        '''A universe with one hydrogen bond acceptor bonding to a hydrogen bond
        donor but in a PBC condition'''
        grofile = '''Test gro file
3
    1ALA      N    1   0.800   0.000   0.000
    1ALA      H    2   0.900   0.000   0.000
    4ALA      O    3   0.100   0.000   0.000
  1.0   1.0   1.0'''
        u = MDAnalysis.Universe(StringIO(grofile), format='gro')
        return u

    @staticmethod
    @pytest.fixture(scope='class')
    def universe_AD():
        '''A universe with one hydrogen bond donor bonding to a hydrogen bond
        acceptor'''
        grofile = '''Test gro file
3
    1ALA      O    1   0.000   0.000   0.000
    4ALA      H    2   0.200   0.000   0.000
    4ALA      N    3   0.300   0.000   0.000
  1.0   1.0   1.0'''
        u = MDAnalysis.Universe(StringIO(grofile), format='gro')
        return u

    @staticmethod
    @pytest.fixture(scope='class')
    def universe_loop():
        '''A universe with one hydrogen bond acceptor bonding to a water which
        bonds back to the first hydrogen bond  acceptor and thus form a loop'''
        grofile = '''Test gro file
5
    1ALA      O    1   0.000   0.001   0.000
    2SOL     OW    2   0.300   0.001   0.000
    2SOL    HW1    3   0.200   0.002   0.000
    2SOL    HW2    4   0.200   0.000   0.000
    4ALA      O    5   0.600   0.000   0.000
  1.0   1.0   1.0'''
        u = MDAnalysis.Universe(StringIO(grofile), format='gro')
        return u

    @staticmethod
    @pytest.fixture(scope='class')
    def universe_DWA():
        '''A universe with one hydrogen bond donor bonding to a hydrogen bond
        acceptor through a water'''
        grofile = '''Test gro file
5
    1ALA      N    1   0.000   0.000   0.000
    1ALA      H    2   0.100   0.000   0.000
    2SOL     OW    3   0.300   0.000   0.000
    2SOL    HW2    4   0.400   0.000   0.000
    4ALA      O    5   0.600   0.000   0.000
  1.0   1.0   1.0'''
        u = MDAnalysis.Universe(StringIO(grofile), format='gro')
        return u

    @staticmethod
    @pytest.fixture(scope='class')
    def universe_DWD():
        '''A universe with one hydrogen bond donor bonding to a hydrogen bond
        donor through a water'''
        grofile = '''Test gro file
5
    1ALA      N    1   0.000   0.000   0.000
    1ALA      H    2   0.100   0.000   0.000
    2SOL     OW    3   0.300   0.000   0.000
    4ALA      H    4   0.500   0.000   0.000
    4ALA      N    5   0.600   0.000   0.000
  1.0   1.0   1.0'''
        u = MDAnalysis.Universe(StringIO(grofile), format='gro')
        return u

    @staticmethod
    @pytest.fixture(scope='class')
    def universe_AWA():
        '''A universe with two hydrogen bond acceptor are joined by a water'''
        grofile = '''Test gro file
5
    1ALA      O    1   0.000   0.000   0.000
    2SOL     OW    2   0.300   0.000   0.000
    2SOL    HW1    3   0.200   0.000   0.000
    2SOL    HW2    4   0.400   0.000   0.000
    4ALA      O    5   0.600   0.000   0.000
  1.0   1.0   1.0'''
        u = MDAnalysis.Universe(StringIO(grofile), format='gro')
        return u

    @staticmethod
    @pytest.fixture(scope='class')
    def universe_AWD():
        '''A universe with one hydrogen bond acceptor bonding to a hydrogen
        bond donor through a water'''
        grofile = '''Test gro file
5
    1ALA      O    1   0.000   0.000   0.000
    2SOL     OW    2   0.300   0.000   0.000
    2SOL    HW1    3   0.200   0.000   0.000
    4ALA      H    4   0.500   0.000   0.000
    4ALA      N    5   0.600   0.000   0.000
  1.0   1.0   1.0'''
        u = MDAnalysis.Universe(StringIO(grofile), format='gro')
        return u

    @staticmethod
    @pytest.fixture(scope='class')
    def universe_AWWA():
        '''A universe with one hydrogen bond acceptor bonding to a hydrogen bond
        acceptor through two waters'''
        grofile = '''Test gro file
7
    1ALA      O    1   0.000   0.000   0.000
    2SOL     OW    2   0.300   0.000   0.000
    2SOL    HW1    3   0.200   0.000   0.000
    2SOL    HW2    4   0.400   0.000   0.000
    3SOL     OW    5   0.600   0.000   0.000
    3SOL    HW1    6   0.700   0.000   0.000
    4ALA      O    7   0.900   0.000   0.000
  1.0   1.0   1.0'''
        u = MDAnalysis.Universe(StringIO(grofile), format='gro')
        return u

    @staticmethod
    @pytest.fixture(scope='class')
    def universe_AWWWA():
        '''A universe with one hydrogen bond acceptor bonding to a hydrogen bond
        acceptor through three waters'''
        grofile = '''Test gro file
9
    1ALA      O    1   0.000   0.000   0.000
    2SOL     OW    2   0.300   0.000   0.000
    2SOL    HW1    3   0.200   0.000   0.000
    2SOL    HW2    4   0.400   0.000   0.000
    3SOL     OW    5   0.600   0.000   0.000
    3SOL    HW1    6   0.700   0.000   0.000
    4SOL     OW    7   0.900   0.000   0.000
    4SOL    HW1    8   1.000   0.000   0.000
    5ALA      O    9   1.200   0.000   0.000
  10.0   10.0   10.0'''
        u = MDAnalysis.Universe(StringIO(grofile), format='gro')
        return u

    @staticmethod
    @pytest.fixture(scope='class')
    def universe_AWWWWA():
        '''A universe with one hydrogen bond acceptor bonding to a hydrogen bond
        acceptor through three waters'''
        grofile = '''Test gro file
11
    1ALA      O    1   0.000   0.000   0.000
    2SOL     OW    2   0.300   0.000   0.000
    2SOL    HW1    3   0.200   0.000   0.000
    2SOL    HW2    4   0.400   0.000   0.000
    3SOL     OW    5   0.600   0.000   0.000
    3SOL    HW1    6   0.700   0.000   0.000
    4SOL     OW    7   0.900   0.000   0.000
    4SOL    HW1    8   1.000   0.000   0.000
    5SOL     OW    9   1.200   0.000   0.000
    5SOL    HW1   10   1.300   0.000   0.000
    6ALA      O   11   1.400   0.000   0.000
  10.0   10.0   10.0'''
        u = MDAnalysis.Universe(StringIO(grofile), format='gro')
        return u

    @staticmethod
    @pytest.fixture(scope='class')
    def universe_branch():
        '''A universe with one hydrogen bond acceptor bonding to two hydrogen
        bond acceptor in selection 2'''
        grofile = '''Test gro file
9
    1ALA      O    1   0.000   0.000   0.000
    2SOL     OW    2   0.300   0.000   0.000
    2SOL    HW1    3   0.200   0.000   0.000
    2SOL    HW2    4   0.400   0.000   0.000
    3SOL     OW    5   0.600   0.000   0.000
    3SOL    HW1    6   0.700   0.000   0.000
    3SOL    HW2    7   0.600   0.100   0.000
    4ALA      O    8   0.900   0.000   0.000
    5ALA      O    9   0.600   0.300   0.000
  1.0   1.0   1.0'''
        u = MDAnalysis.Universe(StringIO(grofile), format='gro')
        return u

    @staticmethod
    @pytest.fixture(scope='class')
    def universe_AWA_AWWA():
        '''A universe with one hydrogen bond acceptors are bonded through one or
        two water'''
        grofile = '''Test gro file
12
    1ALA      O    1   0.000   0.000   0.000
    2SOL     OW    2   0.300   0.000   0.000
    2SOL    HW1    3   0.200   0.000   0.000
    2SOL    HW2    4   0.400   0.000   0.000
    4ALA      O    5   0.600   0.000   0.000
    5ALA      O    6   0.000   1.000   0.000
    6SOL     OW    7   0.300   1.000   0.000
    6SOL    HW1    8   0.200   1.000   0.000
    6SOL    HW2    9   0.400   1.000   0.000
    7SOL     OW   10   0.600   1.000   0.000
    7SOL    HW1   11   0.700   1.000   0.000
    8ALA      O   12   0.900   1.000   0.000
  1.0   1.0   1.0'''
        u = MDAnalysis.Universe(StringIO(grofile), format='gro')
        return u

    @staticmethod
    @pytest.fixture(scope='class')
    def wb_multiframe():
        '''A water bridge object with multipley frames'''
        grofile = '''Test gro file
 13
    1ALA      O    1   0.000   0.000   0.000
    1ALA      H    2   0.000   0.000   0.000
    2SOL     OW    3   0.300   0.000   0.000
    2SOL    HW1    4   0.200   0.000   0.000
    2SOL    HW2    5   0.400   0.000   0.000
    3SOL     OW    6   0.600   0.000   0.000
    3SOL    HW1    7   0.700   0.000   0.000
    4SOL     OW    8   0.900   0.000   0.000
    4SOL    HW1    9   1.000   0.000   0.000
    5SOL     OW   10   1.200   0.000   0.000
    5SOL    HW1   11   1.300   0.000   0.000
    6ALA      H   12   1.400   0.000   0.000
    6ALA      O   13   1.400   0.000   0.000
   10.0   10.0   10.0'''
        u = MDAnalysis.Universe(StringIO(grofile), format='gro')
        wb = WaterBridgeAnalysis(u, 'protein and (resid 1)', 'protein and (resid 4)',
                                 order=4)
        # Build an dummy WaterBridgeAnalysis object for testing
        wb.results.network = []
        wb.results.network.append({(1, 0, 12, None, 2.0, 180.0): None})
        wb.results.network.append({(0, None, 12, 13, 2.0, 180.0): None})
        wb.results.network.append({(1, 0, 3, None, 2.0, 180.0):
                            {(4, 2, 12, None, 2.0, 180.0): None}})
        wb.results.network.append({(0, None, 3, 2, 2.0, 180.0):
                            {(4, 2, 5, None, 2.0, 180.0):
                             {(5, None, 11, 12, 2.0, 180.0): None}}})
        wb.timesteps = range(len(wb.results.network))
        return wb

    def test_nodata(self, universe_DA):
        '''Test if the funtions can run when there is no data.
        This is achieved by not runing the run() first.'''
        wb = WaterBridgeAnalysis(universe_DA, 'protein and (resid 1)',
                                 'protein and (resid 4)', order=0)
        wb.generate_table()
        assert_equal(wb.timesteps_by_type(), None)
        assert_equal(wb.count_by_time(), None)
        assert_equal(wb.count_by_type(), None)

    def test_selection_type_error(self, universe_DA):
        '''Test the case when the wrong selection1_type is given'''
        try:
            wb = WaterBridgeAnalysis(universe_DA, 'protein and (resid 1)',
                'protein and (resid 4)', order=0, selection1_type='aaa')
        except ValueError:
            pass
        else:
            raise pytest.fail("selection_type aaa should rasie error")

    def test_empty_selection(self, universe_DA):
        '''Test the case when selection yields empty result'''
        wb = WaterBridgeAnalysis(universe_DA, 'protein and (resid 9)',
                                 'protein and (resid 10)', order=0)
        wb.run()
        assert wb.results.network == [{}]

    def test_loop(self, universe_loop):
        '''Test if loop can be handled correctly'''
        wb = WaterBridgeAnalysis(universe_loop, 'protein and (resid 1)',
                                 'protein and (resid 1 or resid 4)')
        wb.run()
        assert_equal(len(wb.results.network[0].keys()), 2)

    def test_donor_accepter(self, universe_DA):
        '''Test zeroth order donor to acceptor hydrogen bonding'''
        wb = WaterBridgeAnalysis(universe_DA, 'protein and (resid 1)',
        'protein and (resid 4)', order=0, update_selection=True, debug=True)
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (1, 0, 2, None))

    def test_donor_accepter_heavy(self, universe_DA):
        '''Test zeroth order donor to acceptor hydrogen bonding'''
        wb = WaterBridgeAnalysis(universe_DA, 'protein and (resid 1)',
        'protein and (resid 4)', order=0, update_selection=True, debug=True, distance_type='heavy')
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (1, 0, 2, None))

    def test_donor_accepter_pbc(self, universe_DA_PBC):
        '''Test zeroth order donor to acceptor hydrogen bonding in PBC conditions'''
        wb = WaterBridgeAnalysis(universe_DA_PBC, 'protein and (resid 1)',
            'protein and (resid 4)', order=0, pbc=True)
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (1, 0, 2, None))

    def test_accepter_donor(self, universe_AD):
        '''Test zeroth order acceptor to donor hydrogen bonding'''
        wb = WaterBridgeAnalysis(universe_AD, 'protein and (resid 1)',
                                 'protein and (resid 4)', order=0)
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 1, 2))

    def test_acceptor_water_accepter(self, universe_AWA):
        '''Test case where the hydrogen bond acceptor from selection 1 form
        water bridge with hydrogen bond acceptor from selection 2'''
        wb = WaterBridgeAnalysis(universe_AWA, 'protein and (resid 1)',
                                 'protein and (resid 4)')
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 1, 4, None))
        assert_equal(second[list(second.keys())[0]], None)

    def test_donor_water_accepter(self, universe_DWA):
        '''Test case where the hydrogen bond donor from selection 1 form
        water bridge with hydrogen bond acceptor from selection 2'''
        wb = WaterBridgeAnalysis(universe_DWA, 'protein and (resid 1)',
                                 'protein and (resid 4)')
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (1, 0, 2, None))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 2, 4, None))
        assert_equal(second[list(second.keys())[0]], None)

    def test_acceptor_water_donor(self, universe_AWD):
        '''Test case where the hydrogen bond acceptor from selection 1 form
        water bridge with hydrogen bond donor from selection 2'''
        wb = WaterBridgeAnalysis(universe_AWD, 'protein and (resid 1)',
                                 'protein and (resid 4)')
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (1, None, 3, 4))
        assert_equal(second[list(second.keys())[0]], None)

    def test_donor_water_donor(self, universe_DWD):
        '''Test case where the hydrogen bond donor from selection 1 form
        water bridge with hydrogen bond donor from selection 2'''
        wb = WaterBridgeAnalysis(universe_DWD, 'protein and (resid 1)',
                                 'protein and (resid 4)')
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (1, 0, 2, None))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (2, None, 3, 4))
        assert_equal(second[list(second.keys())[0]], None)

    def test_empty(self, universe_empty):
        '''Test case where no water bridge exists'''
        wb = WaterBridgeAnalysis(universe_empty, 'protein', 'protein')
        wb.run(verbose=False)
        assert_equal(wb.results.network[0], defaultdict(dict))

    def test_same_selection(self, universe_DWA):
        '''
        This test tests that if the selection 1 and selection 2 are both protein.
        However, the protein only forms one hydrogen bond with the water.
        This entry won't be included.
        '''
        wb = WaterBridgeAnalysis(universe_DWA, 'protein and resid 1',
                                 'protein and resid 1')
        wb.run(verbose=False)
        assert_equal(wb.results.network[0], defaultdict(dict))

    def test_acceptor_2water_accepter(self, universe_AWWA):
        '''Test case where the hydrogen bond acceptor from selection 1 form second order
        water bridge with hydrogen bond acceptor from selection 2'''
        # test first order
        wb = WaterBridgeAnalysis(universe_AWWA, 'protein and (resid 1)',
                                 'protein and (resid 4)')
        wb.run(verbose=False)
        assert_equal(wb.results.network[0], defaultdict(dict))
        # test second order
        wb = WaterBridgeAnalysis(universe_AWWA, 'protein and (resid 1)',
                                'protein and (resid 4)', order=2)
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 1, 4, None))
        third = second[list(second.keys())[0]]
        assert_equal(list(third.keys())[0][:4], (5, 4, 6, None))
        assert_equal(third[list(third.keys())[0]], None)
        # test third order
        wb = WaterBridgeAnalysis(universe_AWWA, 'protein and (resid 1)',
                                 'protein and (resid 4)', order=3)
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 1, 4, None))
        third = second[list(second.keys())[0]]
        assert_equal(list(third.keys())[0][:4], (5, 4, 6, None))
        assert_equal(third[list(third.keys())[0]], None)

    def test_acceptor_3water_accepter(self, universe_AWWWA):
        '''Test case where the hydrogen bond acceptor from selection 1 form third order
        water bridge with hydrogen bond acceptor from selection 2'''
        wb = WaterBridgeAnalysis(universe_AWWWA, 'protein and (resid 1)',
                                 'protein and (resid 5)', order=2)
        wb.run(verbose=False)
        assert_equal(wb.results.network[0], defaultdict(dict))

        wb = WaterBridgeAnalysis(universe_AWWWA, 'protein and (resid 1)',
                                 'protein and (resid 5)', order=3)
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 1, 4, None))
        third = second[list(second.keys())[0]]
        assert_equal(list(third.keys())[0][:4], (5, 4, 6, None))
        fourth = third[list(third.keys())[0]]
        assert_equal(list(fourth.keys())[0][:4], (7, 6, 8, None))
        assert_equal(fourth[list(fourth.keys())[0]], None)

        wb = WaterBridgeAnalysis(universe_AWWWA, 'protein and (resid 1)',
                                 'protein and (resid 5)', order=4)
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 1, 4, None))
        third = second[list(second.keys())[0]]
        assert_equal(list(third.keys())[0][:4], (5, 4, 6, None))
        fourth = third[list(third.keys())[0]]
        assert_equal(list(fourth.keys())[0][:4], (7, 6, 8, None))
        assert_equal(fourth[list(fourth.keys())[0]], None)

    def test_acceptor_4water_accepter(self, universe_AWWWWA):
        '''Test case where the hydrogen bond acceptor from selection 1 form fourth order
        water bridge with hydrogen bond acceptor from selection 2'''
        wb = WaterBridgeAnalysis(universe_AWWWWA, 'protein and (resid 1)',
                                 'protein and (resid 6)', order=3)
        wb.run(verbose=False)
        assert_equal(wb.results.network[0], defaultdict(dict))

        wb = WaterBridgeAnalysis(universe_AWWWWA, 'protein and (resid 1)',
                                 'protein and (resid 6)', order=4)
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 1, 4, None))
        third = second[list(second.keys())[0]]
        assert_equal(list(third.keys())[0][:4], (5, 4, 6, None))
        fourth = third[list(third.keys())[0]]
        assert_equal(list(fourth.keys())[0][:4], (7, 6, 8, None))
        fifth = fourth[list(fourth.keys())[0]]
        assert_equal(list(fifth.keys())[0][:4], (9, 8, 10, None))
        assert_equal(fifth[list(fifth.keys())[0]], None)

        wb = WaterBridgeAnalysis(universe_AWWWWA, 'protein and (resid 1)',
                                 'protein and (resid 6)', order=5)
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 1, 4, None))
        third = second[list(second.keys())[0]]
        assert_equal(list(third.keys())[0][:4], (5, 4, 6, None))
        fourth = third[list(third.keys())[0]]
        assert_equal(list(fourth.keys())[0][:4], (7, 6, 8, None))
        fifth = fourth[list(fourth.keys())[0]]
        assert_equal(list(fifth.keys())[0][:4], (9, 8, 10, None))
        assert_equal(fifth[list(fifth.keys())[0]], None)

    def test_acceptor_22water_accepter(self, universe_branch):
        '''Test case where the hydrogen bond acceptor from selection 1 form a second order
        water bridge with hydrogen bond acceptor from selection 2
        and the last water is linked to two residues in selection 2'''
        wb = WaterBridgeAnalysis(universe_branch, 'protein and (resid 1)',
                                 'protein and (resid 4 or resid 5)', order=2)
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 1, 4, None))
        third = second[list(second.keys())[0]]
        assert_equal([(5, 4, 7, None), (6, 4, 8, None)],
                     sorted([key[:4] for key in list(third.keys())]))

    def test_timeseries_wba(self, universe_branch):
        '''Test if the time series data is correctly generated in water bridge analysis format'''
        wb = WaterBridgeAnalysis(universe_branch, 'protein and (resid 1)',
                                 'protein and (resid 4 or resid 5)', order=2)
        wb.output_format = 'sele1_sele2'
        wb.run(verbose=False)
        timeseries = sorted(wb.results.timeseries[0])

        assert_equal(timeseries[0][:4], (0, 2, ('ALA', 1, 'O'), ('SOL', 2, 'HW1')))
        assert_equal(timeseries[1][:4], (3, 4, ('SOL', 2, 'HW2'), ('SOL', 3, 'OW')))
        assert_equal(timeseries[2][:4], (5, 7, ('SOL', 3, 'HW1'), ('ALA', 4, 'O')))
        assert_equal(timeseries[3][:4], (6, 8, ('SOL', 3, 'HW2'), ('ALA', 5, 'O')))

    def test_timeseries_hba(self, universe_branch):
        '''Test if the time series data is correctly generated in hydrogen bond analysis format'''
        wb = WaterBridgeAnalysis(universe_branch, 'protein and (resid 1)',
                                 'protein and (resid 4 or resid 5)', order=2)
        wb.output_format = 'donor_acceptor'
        wb.run(verbose=False)
        timeseries = sorted(wb.results.timeseries[0])

        assert_equal(timeseries[0][:4], (2, 0, ('SOL', 2, 'HW1'), ('ALA', 1, 'O')))
        assert_equal(timeseries[1][:4], (3, 4, ('SOL', 2, 'HW2'), ('SOL', 3, 'OW')))
        assert_equal(timeseries[2][:4], (5, 7, ('SOL', 3, 'HW1'), ('ALA', 4, 'O')))
        assert_equal(timeseries[3][:4], (6, 8, ('SOL', 3, 'HW2'), ('ALA', 5, 'O')))

    def test_acceptor_12water_accepter(self, universe_AWA_AWWA):
        '''Test of independent first order and second can be recognised correctely'''
        wb = WaterBridgeAnalysis(universe_AWA_AWWA, 'protein and (resid 1 or resid 5)',
                                 'protein and (resid 4 or resid 8)', order=1)
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 1, 4, None))
        assert_equal(second[list(second.keys())[0]], None)
        network = wb.results.network[0]
        wb = WaterBridgeAnalysis(universe_AWA_AWWA, 'protein and (resid 1 or resid 5)',
                                 'protein and (resid 4 or resid 8)', order=2)
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal([(0, None, 2, 1), (5, None, 7, 6)],
                     sorted([key[:4] for key in list(network.keys())]))

    def test_count_by_type_single_link(self, universe_DWA):
        '''
        This test tests the simplest water bridge to see if count_by_type() works.
        '''
        wb = WaterBridgeAnalysis(universe_DWA, 'protein and (resid 1)',
                                 'protein and (resid 4)')
        wb.run(verbose=False)
        assert_equal(wb.count_by_type(), [(1, 4, 'ALA', 1, 'H', 'ALA', 4, 'O',  1.)])

    def test_count_by_type_multiple_link(self, universe_AWA_AWWA):
        '''
        This test tests if count_by_type() can give the correct result for more than 1 links.
        '''
        wb = WaterBridgeAnalysis(universe_AWA_AWWA, 'protein and (resid 1 or resid 5)',
                                 'protein and (resid 4 or resid 8)', order=2)
        wb.run(verbose=False)
        assert_equal(sorted(wb.count_by_type()),
        [[0, 4, 'ALA', 1, 'O', 'ALA', 4, 'O', 1.0],
         [5, 11, 'ALA', 5, 'O', 'ALA', 8, 'O', 1.0]])


    def test_count_by_type_multiple_frame(self, wb_multiframe):
        '''
        This test tests if count_by_type() works in multiply situations.
        :return:
        '''
        result = [[0, 11, 'ALA', 1, 'O', 'ALA', 6, 'H', 0.25],
                  [0, 12, 'ALA', 1, 'O', 'ALA', 6, 'O', 0.25],
                  [1, 12, 'ALA', 1, 'H', 'ALA', 6, 'O', 0.5]]

        assert_equal(sorted(wb_multiframe.count_by_type()), result)

    def test_count_by_type_filter(self, wb_multiframe):
        '''
        This test tests if modifying analysis_func
        allows some results to be filtered out in count_by_type().
        :return:
        '''
        def analysis(current, output, u):
            sele1_index, sele1_heavy_index, atom2, heavy_atom2, dist, angle = current[0]
            atom1, heavy_atom1, sele2_index, sele2_heavy_index, dist, angle = current[-1]
            sele1 = u.atoms[sele1_index]
            sele2 = u.atoms[sele2_index]
            (s1_resname, s1_resid, s1_name) = (sele1.resname, sele1.resid, sele1.name)
            (s2_resname, s2_resid, s2_name) = (sele2.resname, sele2.resid, sele2.name)

            key = (sele1_index, sele2_index, s1_resname, s1_resid, s1_name, s2_resname, s2_resid, s2_name)
            if s2_name == 'H':
                output[key] += 1
        result = [((0, 11, 'ALA', 1, 'O', 'ALA', 6, 'H'), 0.25)]
        assert_equal(sorted(wb_multiframe.count_by_type(analysis_func=analysis)), result)

    def test_count_by_type_merge(self, wb_multiframe):
        '''
        This test tests if modifying analysis_func
        allows some same residue to be merged in count_by_type().
        '''
        def analysis(current, output, u):
            sele1_index, sele1_heavy_index, atom2, heavy_atom2, dist, angle = current[0]
            atom1, heavy_atom1, sele2_index, sele2_heavy_index, dist, angle = current[-1]
            sele1 = u.atoms[sele1_index]
            sele2 = u.atoms[sele2_index]
            (s1_resname, s1_resid, s1_name) = (sele1.resname, sele1.resid, sele1.name)
            (s2_resname, s2_resid, s2_name) = (sele2.resname, sele2.resid, sele2.name)

            key = (s1_resname, s1_resid, s2_resname, s2_resid)
            output[key] = 1
        result = [(('ALA', 1, 'ALA', 6), 1.0)]
        assert_equal(sorted(wb_multiframe.count_by_type(analysis_func=analysis)), result)

    def test_count_by_type_order(self, wb_multiframe):
        '''
        This test tests if modifying analysis_func
        allows the order of water bridge to be separated in count_by_type().
        :return:
        '''
        def analysis(current, output, u):
            sele1_index, sele1_heavy_index, atom2, heavy_atom2, dist, angle = current[0]
            atom1, heavy_atom1, sele2_index, sele2_heavy_index, dist, angle = current[-1]
            sele1 = u.atoms[sele1_index]
            sele2 = u.atoms[sele2_index]
            (s1_resname, s1_resid, s1_name) = (sele1.resname, sele1.resid, sele1.name)
            (s2_resname, s2_resid, s2_name) = (sele2.resname, sele2.resid, sele2.name)
            key = (s1_resname, s1_resid, s2_resname, s2_resid, len(current)-1)
            output[key] = 1
        result = [(('ALA', 1, 'ALA', 6, 0), 0.5),
                  (('ALA', 1, 'ALA', 6, 1), 0.25),
                  (('ALA', 1, 'ALA', 6, 2), 0.25)]
        assert_equal(sorted(wb_multiframe.count_by_type(analysis_func=analysis)), result)

    def test_count_by_time(self, wb_multiframe):
        '''
        This test tests if count_by_times() works.
        :return:
        '''
        assert_equal(wb_multiframe.count_by_time(), [(0, 1), (1, 1), (2, 1), (3, 1)])


    def test_count_by_time_weight(self, universe_AWA_AWWA):
        '''
        This test tests if modyfing the analysis_func allows the weight to be changed
        in count_by_type().
        :return:
        '''
        wb = WaterBridgeAnalysis(universe_AWA_AWWA, 'protein and (resid 1 or resid 5)',
                                 'protein and (resid 4 or resid 8)', order=2)
        wb.run(verbose=False)
        def analysis(current, output, u):
            sele1_index, sele1_heavy_index, atom2, heavy_atom2, dist, angle = current[0]
            atom1, heavy_atom1, sele2_index, sele2_heavy_index, dist, angle = current[-1]
            sele1 = u.atoms[sele1_index]
            sele2 = u.atoms[sele2_index]
            (s1_resname, s1_resid, s1_name) = (sele1.resname, sele1.resid, sele1.name)
            (s2_resname, s2_resid, s2_name) = (sele2.resname, sele2.resid, sele2.name)
            key = (s1_resname, s1_resid, s2_resname, s2_resid)
            output[key] += len(current)-1
        assert_equal(wb.count_by_time(analysis_func=analysis), [(0,3), ])

    def test_count_by_time_empty(self, universe_AWA_AWWA):
        '''
        See if count_by_time() can handle zero well.
        :return:
        '''
        wb = WaterBridgeAnalysis(universe_AWA_AWWA, 'protein and (resid 1 or resid 5)',
                                 'protein and (resid 4 or resid 8)', order=2)
        wb.run(verbose=False)
        def analysis(current, output, u):
            pass
        assert_equal(wb.count_by_time(analysis_func=analysis), [(0,0), ])

    def test_generate_table_hba(self, wb_multiframe):
        '''Test generate table using hydrogen bond analysis format'''
        table = wb_multiframe.generate_table(output_format='donor_acceptor')
        assert_array_equal(
            sorted(table.donor_resid),
            [1, 1, 2, 2, 2, 6, 6],
        )

    def test_generate_table_s1s2(self, wb_multiframe):
        '''Test generate table using hydrogen bond analysis format'''
        table = wb_multiframe.generate_table(output_format='sele1_sele2')
        assert_array_equal(
            sorted(table.sele1_resid),
            [1, 1, 1, 1, 2, 2, 3],
        )

    def test_timesteps_by_type(self, wb_multiframe):
        '''Test the timesteps_by_type function'''

        timesteps = sorted(wb_multiframe.timesteps_by_type())
        assert_array_equal(timesteps[3], [1, 12, 'ALA', 1, 'H', 'ALA', 6, 'O', 0, 2])

    def test_duplicate_water(self):
        '''A case #3119 where
        Acceptor···H−O···H-Donor
                     |
                     H···O-H
        will be recognised as 3rd order water bridge.
        '''
        grofile = '''Test gro file
    7
    1LEU      O    1   1.876   0.810   1.354
  117SOL    HW1    2   1.853   0.831   1.162
  117SOL     OW    3   1.877   0.890   1.081
  117SOL    HW2    4   1.908   0.828   1.007
  135SOL     OW    5   1.924   0.713   0.845
    1LEU      H    6   1.997   0.991   1.194
    1LEU      N    7   2.041   1.030   1.274
   2.22092   2.22092   2.22092'''
        u = MDAnalysis.Universe(StringIO(grofile), format='gro')
        wb = WaterBridgeAnalysis(u, 'resname LEU and name O',
                                 'resname LEU and name N H', order=4)
        wb.run()
        assert len(wb.results.timeseries[0]) == 2

    def test_warn_results_deprecated(self, universe_DA):
        wb = WaterBridgeAnalysis(universe_DA, 'protein and (resid 9)',
                                 'protein and (resid 10)', order=0)
        wb.run()

        wmsg = "The `network` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(wb.network, wb.results.network)

        wmsg = "The `timeseries` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(wb.timeseries, wb.results.timeseries)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
from numpy.testing import assert_equal, assert_almost_equal

import MDAnalysis as mda
from MDAnalysis.exceptions import DuplicateWarning, NoDataError
from MDAnalysisTests.datafiles import PSF, DCD, GRO
from MDAnalysisTests.core.util import UnWrapUniverse
import pytest

levels = ('atoms', 'residues', 'segments')

class TestAccumulate(object):
    """Tests the functionality of *Group.accumulate()."""
    @pytest.fixture(params=levels)
    def group(self, request):
        u = mda.Universe(PSF, DCD)
        return getattr(u, request.param)

    def test_accumulate_str_attribute(self, group):
        assert_almost_equal(group.accumulate("masses"), np.sum(group.atoms.masses))

    def test_accumulate_different_func(self, group):
        assert_almost_equal(
            group.accumulate("masses", function=np.prod),
            np.prod(group.atoms.masses))

    @pytest.mark.parametrize('name, compound', (('resindices', 'residues'),
                                                ('segindices', 'segments'),
                                                ('molnums', 'molecules'),
                                                ('fragindices', 'fragments')))
    @pytest.mark.parametrize('level', levels)
    def test_accumulate_str_attribute_compounds(self, name, compound, level):
        u = UnWrapUniverse()
        group = getattr(u, level)
        ref = [sum(a.masses) for a in group.atoms.groupby(name).values()]
        vals = group.accumulate("masses", compound=compound)
        assert_almost_equal(vals, ref, decimal=5)

    def test_accumulate_wrongname(self, group):
        with pytest.raises(AttributeError):
            group.accumulate("foo")

    def test_accumulate_wrongcomponent(self, group):
        with pytest.raises(ValueError):
            group.accumulate("masses", compound="foo")

    @pytest.mark.parametrize('level', levels)
    def test_accumulate_nobonds(self, level):
        group = getattr(mda.Universe(GRO), level)
        with pytest.raises(NoDataError):
            group.accumulate("masses", compound="fragments")

    @pytest.mark.parametrize('level', levels)
    def test_accumulate_nomolnums(self, level):
        group = getattr(mda.Universe(GRO), level)
        with pytest.raises(NoDataError):
            group.accumulate("masses", compound="molecules")

    def test_accumulate_array_attribute(self, group):
        a = np.ones((len(group.atoms), 2, 5))
        assert_equal(group.accumulate(a), np.sum(a, axis=0))

    def test_accumulate_array_attribute_wrongshape(self, group):
        with pytest.raises(ValueError):
            group.accumulate(np.ones(len(group.atoms) - 1))

    @pytest.mark.parametrize('name, compound', (('resindices', 'residues'),
                                                ('segindices', 'segments'),
                                                ('molnums', 'molecules'),
                                                ('fragindices', 'fragments')))
    @pytest.mark.parametrize('level', levels)
    def test_accumulate_array_attribute_compounds(self, name, compound, level):
        u = UnWrapUniverse()
        group = getattr(u, level)
        ref = [np.ones((len(a), 2, 5)).sum(axis=0) for a in group.atoms.groupby(name).values()]
        assert_equal(group.accumulate(np.ones((len(group.atoms), 2, 5)), compound=compound), ref)

class TestTotals(object):
    """Tests the functionality of *Group.total*() like total_mass
    and total_charge.
    """
    @pytest.fixture(params=levels)
    def group(self, request):
        u = mda.Universe(PSF, DCD)
        return getattr(u, request.param)

    def test_total_charge(self, group):
        assert_almost_equal(group.total_charge(), -4.0, decimal=4)

    @pytest.mark.parametrize('name, compound',
                             (('resids', 'residues'), ('segids', 'segments'),
                              ('fragindices', 'fragments')))
    def test_total_charge_compounds(self, group, name, compound):
        ref = [sum(a.charges) for a in group.atoms.groupby(name).values()]
        assert_almost_equal(group.total_charge(compound=compound), ref)

    @pytest.mark.filterwarnings(  # Prevents regression of issue #2990
        "error:"
        "Using a non-tuple sequence for multidimensional indexing is deprecated:"
        "FutureWarning"
    )
    def test_total_charge_duplicates(self, group):
        group2 = group + group[0]
        ref = group.total_charge() + group[0].charge
        with pytest.warns(DuplicateWarning) as w:
            assert_almost_equal(group2.total_charge(), ref)
            assert len(w) == 1

    def test_total_mass(self, group):
        assert_almost_equal(group.total_mass(), 23582.043)

    @pytest.mark.parametrize('name, compound',
                             (('resids', 'residues'), ('segids', 'segments'),
                              ('fragindices', 'fragments')))
    def test_total_mass_compounds(self, group, name, compound):
        ref = [sum(a.masses) for a in group.atoms.groupby(name).values()]
        assert_almost_equal(group.total_mass(compound=compound), ref)

    @pytest.mark.filterwarnings(  # Prevents regression of issue #2990
        "error:"
        "Using a non-tuple sequence for multidimensional indexing is deprecated:"
        "FutureWarning"
    )
    def test_total_mass_duplicates(self, group):
        group2 = group + group[0]
        ref = group.total_mass() + group2[0].mass
        with pytest.warns(DuplicateWarning) as w:
            assert_almost_equal(group2.total_mass(), ref)
            assert len(w) == 1



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import MDAnalysis
import pytest

from MDAnalysis.tests.datafiles import (
    CRD,
    LAMMPSdata,
    DLP_CONFIG_minimal,
    GRO,
    TPR,
    DMS,
    GMS_SYMOPT,
    MMTF,
    mol2_molecule,
    PRM7,
    PDB_small,
    PDBQT_input,
    PQR,
    PRM,
    PSF,
    PRM12,
    HoomdXMLdata,
    XPDB_small,
    XYZ_mini,
    DLP_HISTORY_minimal, )


@pytest.mark.parametrize('prop', [
    'name',
    'resname',
    'type',
    'segid',
    'moltype',
])
# topology formats curated from values available in
# MDAnalysis._PARSERS
@pytest.mark.parametrize( 'top_format, top', [
    ('CONFIG', DLP_CONFIG_minimal),
    ('CRD', CRD),
    ('DATA', LAMMPSdata),
    ('DMS', DMS),
    ('GMS', GMS_SYMOPT),
    ('GRO', GRO),
    ('HISTORY', DLP_HISTORY_minimal),
    ('MMTF', MMTF),
    ('MOL2', mol2_molecule),
    ('PARM7', PRM7),
    ('PDB', PDB_small),
    ('PDBQT', PDBQT_input),
    ('PQR', PQR),
    ('PRMTOP', PRM),
    ('PSF', PSF),
    ('TOP', PRM12),
    ('TPR', TPR),
    ('XML', HoomdXMLdata),
    ('XPDB', XPDB_small),
    ('XYZ', XYZ_mini)
])
def test_str_types(top_format, top, prop):
    # Python 2/3 topology string type checking
    # Related to Issue #1336
    u = MDAnalysis.Universe(top, format=top_format)
    if hasattr(u.atoms[0], prop):
        assert isinstance(getattr(u.atoms[0], prop), str)
